#!/usr/bin/env python3

import argparse

from bson.min_key import MinKey
from bson.son import SON
from pymongo.errors import OperationFailure

try:
    from pymongo import MongoClient as Connection
except ImportError:
    from pymongo import Connection


def presplit(host, database, collection, shardkey, shardnumber=None,
             chunkspershard=1, verbose=False):
    """
    Presplit chunks for sharding.

    Get information about the number of shards, then split chunks and
    distribute over shards. Currently assumes shardkey to be hex string, for
    example ObjectId or UUID.

    host: host and port to connect to, e.g. "192.168.0.1:27017",
          "localhost:30000"
    database: database name to enable sharding
    collection: collection name to shard
    shardkey: shardkey to pre-split on (must be hex string, e.g. ObjectId or
              UUID)
    shardnumber: if None, automatically presplit over all available shards.
                 if integer, only presplit over the given number of shards
                 (maximum is the number of actual shards)
    """
    con = Connection(host)
    namespace = '%s.%s' % (database, collection)

    # disable balancer
    con['config']['settings'].update_one({'_id': "balancer"},
                                     {'$set': {'stopped': True}}, upsert=True)

    # enable sharding on database if not yet enabled
    db_info = con['config']['databases'].find_one({'_id': database})
    if not db_info or db_info['partitioned'] is False:
        con['admin'].command(SON({'enableSharding': database}))

    # shard collection if not yet sharded
    coll_info = con['config']['collections'].find_one({'_id': namespace})
    if coll_info and not coll_info['dropped']:
        # if it is sharded already, quit. something is not right.
        if verbose:
            print("collection already sharded.")
        return
    else:
        con[database][collection].create_indexes([shardkey])
        con['admin'].command(SON({'shardCollection': namespace,
                                  'key': {shardkey: 1}}))

    # get shard number and names and calculate split points
    shards = list(con['config']['shards'].find())

    if len(shards) == 1:
        if verbose:
            print("only one shard found. no pre-splitting required.")
        return

    # limit number of shards if shardnumber given
    if shardnumber and shardnumber <= len(shards):
        shards = shards[:shardnumber]

    shard_names = [s['_id'] for s in shards]
    splits_total = len(shards) * chunkspershard
    split_interval = 16**4 / splits_total
    split_points = ["%0.4x" % s for s in range(split_interval,
                                               splits_total * split_interval,
                                               split_interval)]

    # pre-splitting commands
    for s in split_points:
        con['admin'].command(SON([('split', namespace),
                                  ('middle', {shardkey: s})]))

    split_points = [MinKey()] + split_points

    # move chunks to shards (catch the one error where the chunk resides
    # on that shard already)
    for i, s in enumerate(split_points):
        try:
            if verbose:
                print('moving chunk %s in collection %s to shard %s.'
                      % (s, namespace, shard_names[i % len(shards)]))
            res = con['admin'].command(SON([('moveChunk', namespace),
                                            ('find', {shardkey: s}),
                                            ('to',
                                             shard_names[i % len(shards)])]))
        except OperationFailure as e:
            if verbose:
                print(e)

    if verbose:
        print('chunk distribution:', end=' ')
        chunk_group = con['config']['chunks'].aggregate([
            { '$match': { 'ns': namespace }},
            { '$group': { 
                '_id': '$shard',
                'nChunks': { '$sum': 1 }}
            }
        ])
        print(', '.join(["%s: %i" % (ch['_id'], ch['nChunks'])
                         for ch in chunk_group]))


if __name__ == '__main__':

    # presplitting function
    parser = argparse.ArgumentParser(description='MongoDB pre-splitting tool')

    parser.add_argument('host', action='store', nargs='?',
                        default='localhost:27017', metavar='host:port',
                        help=('host:port of mongos or mongod process '
                              '(default localhost:27017)'))
    parser.add_argument('namespace', action='store',
                        help=('namespace to shard, '
                              'in form "database.collection"'))
    parser.add_argument('shardkey', action='store',
                        help='shard key to split on, e.g. "_id"')
    parser.add_argument('-n', '--number', action='store', metavar='N',
                        type=int, default=None,
                        help='max. number of shards to use (default is all)')
    parser.add_argument('-c', '--chunks', action='store', metavar='N',
                        type=int, default=1,
                        help='number of chunks per shard (default is 1)')

    parser.add_argument('--verbose', action='store_true', default=False,
                        help='print verbose information')
    args = vars(parser.parse_args())

    args['database'], args['collection'] = args['namespace'].split('.')
    presplit(args['host'], args['database'], args['collection'],
             args['shardkey'], args['number'], args['chunks'], args['verbose'])



import re
from collections import namedtuple
from operator import itemgetter
from mtools.util import OrderedDict
from mtools.util.grouping import Grouping
from mtools.util.print_table import print_table
from .base_section import BaseSection

LogTuple = namedtuple('LogTuple', ['datetime', 'txnNumber', 'autocommit', 'readConcern',
                                   'timeActiveMicros', 'timeInactiveMicros', 'duration'])


def op_or_cmd(le):
    return le.operation if le.operation != 'command' else le.command


class TransactionSection(BaseSection):
    """TransactionSection class."""

    name = "transactions"

    def __init__(self, mloginfo):
        BaseSection.__init__(self, mloginfo)

        # add --transactions flag to argparser
        helptext = 'outputs statistics about transactions'
        self.mloginfo.argparser_sectiongroup.add_argument('--transactions',
                                                          action='store_true',
                                                          help=helptext)

        # add --tsort flag to argparser for transaction sort
        self.mloginfo.argparser_sectiongroup.add_argument('--tsort',
                                                          action='store',

                                                          choices=['duration'])

    @property
    def active(self):
        """Return boolean if this section is active."""
        return self.mloginfo.args['transactions']

    def run(self):

        """Run this section and print out information."""
        grouping = Grouping(group_by=lambda x: (x.datetime, x.txnNumber,
                                                x.autocommit, x.readConcern,
                                                x.timeActiveMicros,
                                                x.timeInactiveMicros,
                                                x.duration))

        logfile = self.mloginfo.logfile

        if logfile.start and logfile.end:
            progress_start = self.mloginfo._datetime_to_epoch(logfile.start)
            progress_total = (self.mloginfo._datetime_to_epoch(logfile.end) -
                              progress_start)
        else:
            self.mloginfo.progress_bar_enabled = False

        for i, le in enumerate(logfile):
            # update progress bar every 1000 lines

            if self.mloginfo.progress_bar_enabled and (i % 1000 == 0):
                if le.datetime:
                    progress_curr = self.mloginfo._datetime_to_epoch(le.datetime)
                    if progress_total:
                        (self.mloginfo.update_progress(
                            float(progress_curr - progress_start) / progress_total))

            if re.search('transaction', le.line_str):
                lt = LogTuple(le.datetime, le.txnNumber, le.autocommit, le.readConcern,
                              le.timeActiveMicros, le.timeInactiveMicros, le.duration)

                grouping.add(lt)

        grouping.sort_by_size()

        # clear progress bar again
        if self.mloginfo.progress_bar_enabled:
            self.mloginfo.update_progress(1.0)

        # no queries in the log file
        if not len(grouping):
            print('no transactions found.')
            return

        titles = ['datetime', 'txnNumber', 'autocommit', 'readConcern',
                  'timeActiveMicros', 'timeInactiveMicros', 'duration']

        table_rows = []
        # using only important key-values
        # can be used in future
        for g in grouping:
            # calculate statistics for this group
            (datetime, txnNumber, autocommit, readConcern, timeActiveMicros,
             timeInactiveMicros, duration) = g

            stats = OrderedDict()
            stats['datetime'] = str(datetime)
            stats['txnNumber'] = txnNumber
            stats['autocommit'] = autocommit
            stats['readConcern'] = readConcern
            stats['timeActiveMicros'] = timeActiveMicros
            stats['timeInactiveMicros'] = timeInactiveMicros
            stats['duration'] = duration
            table_rows.append(stats)

        if self.mloginfo.args['tsort'] == 'duration':
            table_rows = sorted(table_rows,
                                key=itemgetter(self.mloginfo.args['tsort']),
                                reverse=True)

        print_table(table_rows, titles, uppercase_headers=True)

        print('')



#!/usr/bin/env python3

import os
import re
# import sys
# import argparse
from itertools import zip_longest
import pickle

import mtools


def import_l2c_db():
    """
    Static import helper function.

    Checks if the log2code.pickle exists first, otherwise raises ImportError.
    """
    data_path = os.path.join(os.path.dirname(mtools.__file__), 'data')
    if os.path.exists(os.path.join(data_path, 'log2code.pickle')):
        av, lv, lbw, lcl = pickle.load(open(os.path.join(data_path,
                                                          'log2code.pickle'),
                                             'rb'))
        return av, lv, lbw, lcl
    else:

        raise ImportError('log2code.pickle not found in %s.' % data_path)


class Log2CodeConverter(object):

    # static import of logdb data structures
    all_versions, log_version, logs_by_word, log_code_lines = import_l2c_db()

    def _log2code(self, line):
        tokens = re.split(r'[\s"]', line)

        # find first word in first 20 tokens that has a corresponding
        # log message stored
        for word_no, word in enumerate(w for w in tokens
                                       if w in self.logs_by_word):

            # go through all error messages starting with this word
            coverage = []
            for log in self.logs_by_word[word]:

                if all([line.find(token) >= 0 for token in log]):
                    # all tokens match, calculate coverage
                    cov = sum([len(token) for token in log])
                    coverage.append(cov)
                else:
                    coverage.append(0)

            best_cov = max(coverage)
            if not best_cov:
                continue

            if word_no > 20:
                # avoid parsing really long lines. If the log message didn't
                # start within the first 20 words, it's probably not a
                # known message
                return None

                # no match found, may have been a named log level. try next
                # word.
                # if word in ["warning:", "ERROR:", "SEVERE:", "UNKNOWN:"]:
                #     continue
                # else:
                #     # duration = time.time() - start_time
                #     # print(duration)
                #     continue

            best_match = self.logs_by_word[word][coverage.index(best_cov)]
            return self.log_code_lines[best_match]

    def _strip_counters(self, sub_line):
        """Find the codeline end by taking out the counters and durations."""
        try:
            end = sub_line.rindex('}')
        except ValueError:
            return sub_line
        else:
            return sub_line[:(end + 1)]

    def _strip_datetime(self, sub_line):
        """Strip datetime and other parts so that there is no redundancy."""
        try:
            begin = sub_line.index(']')
        except ValueError:
            return sub_line
        else:
            # create a "" in place character for the beginnings..
            # needed when interleaving the lists
            sub = sub_line[begin + 1:]
            return sub

    def _find_variable(self, pattern, logline):
        """
        Return the variable parts of the code given a tuple of strings pattern.

        Example: (this, is, a, pattern) -> 'this is a good pattern' -> [good]
        """
        var_subs = []
        # find the beginning of the pattern
        first_index = logline.index(pattern[0])
        beg_str = logline[:first_index]
        # strip the beginning substring
        var_subs.append(self._strip_datetime(beg_str))

        for patt, patt_next in zip(pattern[:-1], pattern[1:]):
            # regular expression pattern that finds what's in the middle of
            # two substrings
            pat = re.escape(patt) + '(.*)' + re.escape(patt_next)
            # extract whats in the middle of the two substrings
            between = re.search(pat, logline)
            try:
                # add what's in between if the search isn't none
                var_subs.append(between.group(1))
            except Exception:
                pass
        rest_of_string = logline.rindex(pattern[-1]) + len(pattern[-1])

        # add the rest of the string to end minus the counters and durations
        end_str = logline[rest_of_string:]
        var_subs.append(self._strip_counters(end_str))

        # strip whitespace from each string, but keep the strings themselves
        # var_subs = [v.strip() for v in var_subs]

        return var_subs

    def _variable_parts(self, line, codeline):
        """Return variable parts of the codeline, given the static parts."""
        var_subs = []
        # codeline has pattern and then has the outputs in different versions
        if codeline:
            var_subs = self._find_variable(codeline.pattern, line)
        else:
            # make variable part of the line string without all the other stuff
            line_str = self._strip_datetime(self._strip_counters(line))
            var_subs = [line_str.strip()]
        return var_subs

    def __call__(self, line, variable=False):
        """Return tuple of log2code and variable parts when class is called."""
        if variable:
            log2code = self._log2code(line)
            return log2code, self._variable_parts(line, log2code)
        else:
            return self._log2code(line), None

    def combine(self, pattern, variable):
        """Combine a pattern and variable parts to be a line string again."""
        inter_zip = zip_longest(variable, pattern, fillvalue='')
        interleaved = [elt for pair in inter_zip for elt in pair]
        return ''.join(interleaved)


# class MLog2Code(object):
#
#     def __init__(self):
#         self._import_l2c_db()
#         self._parse_args()
#         self.analyse()
#
#     def _import_l2c_db(self):
#         (self.all_versions, self.logs_versions, self.logs_by_word,
#          self.log_code_lines) = cPickle.load(open('./logdb.pickle', 'rb'))
#
#     def _parse_args(self):
#         # create parser object
#         parser = argparse.ArgumentParser(description=('mongod/mongos log '
#                                                       'file to code line '
#                                                       'converter (BETA)'))
#
#         # only create default argument if not using stdin
#         if sys.stdin.isatty():
#             parser.add_argument('logfile', action='store',
#                                 help=('looks up and prints out information '
#                                       'about where a log line originates '
#                                       'from the code.'))
#
#         self.args = vars(parser.parse_args())
#
#     def analyse(self):
#         # open logfile
#         if sys.stdin.isatty():
#             logfile = open(self.args['logfile'], 'r')
#         else:
#             logfile = sys.stdin
#
#         for i, line in enumerate(logfile):
#             match = self.log2code(line)
#
#             if  match:
#                 print(line, end=' ')
#                 print(self.logs_versions[match])
#                 print(self.log_code_lines[match])
#
#
#     def log2code(self, line):
#         tokens = line.split()
#
#         # find first word in line that has a corresponding log message stored
#         word = next((w for w in tokens if w in self.logs_by_word), None)
#         if not word:
#             return None
#
#         # go through all error messages starting with this word
#         coverage = []
#         for log in self.logs_by_word[word]:
#
#             if all([line.find(token) >= 0 for token in log]):
#                 # all tokens match, calculate coverage
#                 cov = sum([len(token) for token in log])
#                 coverage.append(cov)
#             else:
#                 coverage.append(0)
#
#         best_cov = max(coverage)
#         if not best_cov:
#             return None
#
#         best_match = self.logs_by_word[word][coverage.index(best_cov)]
#         return best_match
#
#
# if __name__ == '__main__':
#     l2cc = Log2CodeConverter()
#     lcl = l2cc('Sun Mar 24 00:44:16.295 [conn7815] moveChunk migrate '
#                'commit accepted by TO-shard: '
#                '{ active: true, ns: "db.coll", from: "shard001:27017", '
#                'min: { i: ObjectId("4b7730748156791f310b03a3"), '
#                'm: "stats", t: new Date(1348272000000) }, '
#                'max: { i: ObjectId("4b8f826192f9e2154d05dda7"), '
#                'm: "mongo", t: new Date(1345680000000) }, '
#                'shardKeyPattern: { i: 1.0, m: 1.0, t: 1.0 }, '
#                'state: "done", counts: { cloned: 3115, '
#                'clonedBytes: 35915282, catchup: 0, '
#                'steady: 0 }, ok: 1.0 }')
#     print lcl.versions
#
#     possible_versions = possible_versions & set(logs_versions[best_match])
#
#     if len(possible_versions) != old_num_v:
#         print(i, line.rstrip())
#         print("    best_match: %s" % best_match)
#         print("    log message only present in versions: %s"
#               % logs_versions[best_match])
#         print("    this limits the possible versions to: %s"
#               % possible_versions)
#         print('')
#
#     if not possible_versions:
#         raise SystemExit
#
#     print("possible versions: "
#           ", ".join([pv[1:] for pv in possible_versions]))
#     for pv in possible_versions:
#         print(pv, possible_versions[pv])
#
#     plt.bar(range(len(possible_versions.values())),
#             possible_versions.values(), align='center')
#     plt.xticks(range(len(possible_versions.keys())),
#                possible_versions.keys(), size='small', rotation=90)
#     plt.show()


import inspect
import json
import os
import shutil
import sys
import time
import unittest

from bson import SON

import pytest
from pymongo import MongoClient
from pymongo.errors import ConnectionFailure

from mtools.mlaunch.mlaunch import MLaunchTool

pytestmark = pytest.mark.skip("skip all mlaunch tests for now")

class TestMLaunch(object):
    """
    This class tests functionality around the mlaunch tool. It has some
    additional methods that are helpful for the tests, as well as a setup
    and teardown method for all tests.

    Don't call tests from other tests. This won't work as each test gets
    its own data directory (for debugging).
    """

    port = 33333
    base_dir = 'data_test_mlaunch'

    def __init__(self):
        """Constructor."""
        self.use_auth = False
        self.data_dir = ''

    def setup_method(self):
        """Start up method to create mlaunch tool and find free port."""
        self.tool = MLaunchTool(test=True)

        # if the test data path exists, remove it
        if os.path.exists(self.base_dir):
            shutil.rmtree(self.base_dir)

    def teardown_method(self):
        """Tear down method after each test, removes data directory."""

        # kill all running processes
        self.tool.discover()

        ports = self.tool.get_tagged(['all', 'running'])
        processes = self.tool._get_processes().values()
        for p in processes:
            p.terminate()
            p.wait(10)

        self.tool.wait_for(ports, to_start=False)

        # quick sleep to avoid spurious test failures
        time.sleep(1)

        # if the test data path exists, remove it
        if os.path.exists(self.base_dir):
            shutil.rmtree(self.base_dir)

    def run_tool(self, arg_str):
        """Wrapper to call self.tool.run() with or without auth."""
        # name data directory according to test method name
        caller = inspect.stack()[1][3]
        self.data_dir = os.path.join(self.base_dir, caller)

        # add data directory to arguments for all commands
        arg_str += ' --dir %s' % self.data_dir

        if arg_str.startswith('init') or arg_str.startswith('--'):
            # add --port and --nojournal to init calls
            arg_str += ' --port %i --nojournal --smallfiles' % self.port

            if self.use_auth:
                # add --auth to init calls if flag is set
                arg_str += ' --auth'

        self.tool.run(arg_str)

    # -- tests below ---

    @pytest.mark.xfail(raises=ConnectionFailure)
    def test_test(self):
        """TestMLaunch setup and teardown test."""

        # test that data dir does not exist
        assert not os.path.exists(self.data_dir)

        # start mongo process on free test port
        self.run_tool("init --single")

        # call teardown method within this test
        self.teardown()

        # test that data dir does not exist anymore
        assert not os.path.exists(self.data_dir)

        # test that mongod is not running on this port anymore
        # (raises ConnectionFailure)
        mc = MongoClient('localhost:%i' % self.port,
                         serverSelectionTimeoutMS=100).server_info()
        print(mc['version'])

    def test_argv_run(self):
        """
        mlaunch: test true command line arguments, instead of passing
        into tool.run().
        """

        # make command line arguments through sys.argv
        sys.argv = ['mlaunch', 'init', '--single', '--dir', self.base_dir,
                    '--port', str(self.port), '--nojournal']

        self.tool.run()
        assert self.tool.is_running(self.port)

    def test_init_default(self):
        """mlaunch: test that 'init' command can be omitted, is default. """

        # make command line arguments through sys.argv
        sys.argv = ['mlaunch', '--single', '--dir', self.base_dir,
                    '--port', str(self.port), '--nojournal']

        self.tool.run()
        assert self.tool.is_running(self.port)

    def test_init_default_arguments(self):
        """
        mlaunch: test that 'init' command is default, even when specifying
        arguments to run().
        """

        self.run_tool("--single")
        assert self.tool.is_running(self.port)

    def test_single(self):
        """mlaunch: start stand-alone server and tear down again."""

        # start mongo process on free test port
        self.run_tool("init --single")

        # make sure node is running
        assert self.tool.is_running(self.port)

        # check if data directory and logfile exist
        assert os.path.exists(os.path.join(self.data_dir, 'db'))
        assert os.path.isfile(os.path.join(self.data_dir, 'mongod.log'))

        # check that the tags are set correctly: 'single', 'mongod',
        # 'running', <port>
        assert set(self.tool.get_tags_of_port(self.port)) == set(['running',
                                                                  'mongod',
                                                                  'all',
                                                                  'single',
                                                                  str(self.
                                                                      port)])

    def test_replicaset_conf(self):
        """Start replica set of 2 nodes + arbiter and compare rs.conf()."""

        # start mongo process on free test port
        self.run_tool("init --replicaset --nodes 2 --arbiter")

        # check if data directories exist
        assert os.path.exists(os.path.join(self.data_dir, 'replset'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs1'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs2'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/arb'))

        # create mongo client for the next tests
        mc = MongoClient('localhost:%i' % self.port)

        # get rs.conf() and check for 3 members, exactly one is arbiter
        conf = mc['local']['system.replset'].find_one()
        assert len(conf['members']) == 3
        assert sum(1 for memb in conf['members']
                   if 'arbiterOnly' in memb and memb['arbiterOnly']) == 1

    @pytest.mark.timeout(60)
    @pytest.mark.slow
    def test_replicaset_ismaster(self):
        """Start replica set and verify that first node becomes primary."""

        # start mongo process on free test port
        self.run_tool("init --replicaset")

        # wait for primary
        assert self.tool._wait_for_primary()

        # insert a document and wait to replicate to 2 secondaries
        # (10 sec timeout)
        mc = MongoClient('localhost:%i' % self.port)
        mc.test.smokeWait.insert_one({}, w=2, wtimeout=10 * 60 * 1000)

    @unittest.skip('incompatible with 3.4 CSRS')
    def test_sharded_status(self):
        """Start cluster with 2 shards of single nodes, 1 config server."""

        # start mongo process on free test port
        self.run_tool("init --sharded 2 --single")

        # check if data directories and logfile exist
        assert os.path.exists(os.path.join(self.data_dir, 'shard01/db'))
        assert os.path.exists(os.path.join(self.data_dir, 'shard02/db'))
        assert os.path.exists(os.path.join(self.data_dir, 'config/db'))
        assert os.path.isfile(os.path.join(self.data_dir, 'mongos.log'))

        # create mongo client
        mc = MongoClient('localhost:%i' % (self.port))

        # check for 2 shards and 1 mongos
        assert mc['config']['shards'].count_documents({}) == 2
        assert mc['config']['mongos'].count_documents({}) == 1

    def helper_output_has_line_with(self, keywords, output):
        """Check if output contains a line where all keywords are present."""
        return len(filter(None, [all([kw in line for kw in keywords])
                                 for line in output]))

    @unittest.skip('incompatible with 3.4 CSRS')
    def test_verbose_sharded(self):
        """Test verbose output when creating sharded cluster."""

        self.run_tool("init --sharded 2 --replicaset --config 3 "
                      "--mongos 2 --verbose")

        # capture stdout
        output = sys.stdout.getvalue().splitlines()

        keywords = ('rs1', 'rs2', 'rs3', 'shard01', 'shard02', 'config1',
                    'config2', 'config3')

        # creating directory
        for keyword in keywords:
            # make sure every directory creation was announced to stdout
            assert self.helper_output_has_line_with(['creating directory',
                                                     keyword, 'db'], output)

        assert self.helper_output_has_line_with(['creating directory',
                                                 'mongos'], output)

        # launching nodes
        for keyword in keywords:
            assert self.helper_output_has_line_with(['launching', keyword,
                                                     '--port', '--logpath',
                                                     '--dbpath'], output)

        # mongos
        assert self.helper_output_has_line_with(['launching', 'mongos',
                                                 '--port', '--logpath',
                                                 str(self.port)], output)
        assert self.helper_output_has_line_with(['launching', 'mongos',
                                                 '--port', '--logpath',
                                                 str(self.port + 1)], output)

        # some fixed outputs
        assert self.helper_output_has_line_with(['waiting for nodes to '
                                                 'start'], output)
        assert self.helper_output_has_line_with(['adding shards. can take up '
                                                 'to 30 seconds'], output)
        assert self.helper_output_has_line_with(['writing .mlaunch_startup '
                                                 'file'], output)
        assert self.helper_output_has_line_with(['done'], output)

        # replica sets initialized, shard added
        for keyword in ('shard01', 'shard02'):
            assert self.helper_output_has_line_with(['replica set', keyword,
                                                     'initialized'], output)
            assert self.helper_output_has_line_with(['shard', keyword,
                                                     'added successfully'],
                                                    output)

    def test_shard_names(self):
        """mlaunch: test if sharded cluster with explicit shard names works."""

        # start mongo process on free test port
        self.run_tool("init --sharded tic tac toe --replicaset")

        # create mongo client
        mc = MongoClient('localhost:%i' % (self.port))

        # check that shard names match
        shard_names = set(doc['_id'] for doc in mc['config']['shards'].find())
        assert shard_names == set(['tic', 'tac', 'toe'])

    def test_startup_file(self):
        """mlaunch: create .mlaunch_startup file in data path."""

        # Also tests utf-8 to byte conversion and json import

        self.run_tool("init --single -v")

        # check if the startup file exists
        startup_file = os.path.join(self.data_dir, '.mlaunch_startup')
        assert os.path.isfile(startup_file)

        # compare content of startup file with tool.args
        file_contents = json.load(open(startup_file, 'rb'))
        assert file_contents['parsed_args'] == self.tool.args
        assert file_contents['unknown_args'] == self.tool.unknown_args

    def test_single_mongos_explicit(self):
        """
        mlaunch: test if single mongos is running on start port and creates
        <datadir>/mongos.log.
        """

        # start 2 shards, 1 config server, 1 mongos
        self.run_tool("init --sharded 2 --single --config 1 --mongos 1")

        # check if mongos log files exist on correct ports
        assert os.path.exists(os.path.join(self.data_dir, 'mongos.log'))

        # check for correct port
        assert self.tool.get_tagged('mongos') == set([self.port])

    def test_single_mongos(self):
        """
        mlaunch: test if multiple mongos use separate log files in 'mongos'
        subdir.
        """

        # start 2 shards, 1 config server, 2 mongos
        self.run_tool("init --sharded 2 --single --config 1 --mongos 1")

        # check that 2 mongos are running
        assert len(self.tool.get_tagged(['mongos', 'running'])) == 1

    def test_multiple_mongos(self):
        """
        mlaunch: test if multiple mongos use separate log files in 'mongos'
        subdir.
        """

        # start 2 shards, 1 config server, 2 mongos
        self.run_tool("init --sharded 2 --single --config 1 --mongos 2")

        # this also tests that mongos are started at the beginning of the
        # port range
        assert os.path.exists(os.path.join(self.data_dir, 'mongos',
                                           'mongos_%i.log' % (self.port)))
        assert os.path.exists(os.path.join(self.data_dir, 'mongos',
                                           'mongos_%i.log' % (self.port + 1)))

        # check that 2 mongos are running
        assert len(self.tool.get_tagged(['mongos', 'running'])) == 2

    def test_filter_valid_arguments(self):
        """Check arguments unknown to mlaunch against mongos and mongod."""

        # filter against mongod
        result = self.tool._filter_valid_arguments("--slowms 500 -vvv "
                                                   "--configdb localhost:27017"
                                                   " --foobar".split(),
                                                   "mongod")
        assert result == "--slowms 500 -vvv"

        # filter against mongos
        result = self.tool._filter_valid_arguments("--slowms 500 -vvv "
                                                   "--configdb localhost:27017"
                                                   " --foobar".split(),
                                                   "mongos")
        assert result == "-vvv --configdb localhost:27017"

    def test_large_replicaset_arbiter(self):
        """mlaunch: start large replica set of 7 nodes with arbiter."""

        # start mongo process on free test port
        # (don't need journal for this test)
        self.run_tool("init --replicaset --nodes 6 --arbiter")

        # check if data directories exist
        assert os.path.exists(os.path.join(self.data_dir, 'replset'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs1'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs2'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs3'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs4'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs5'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs6'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/arb'))

        # create mongo client for the next tests
        mc = MongoClient('localhost:%i' % self.port)

        # get rs.conf() and check for 7 members, exactly one arbiter
        conf = mc['local']['system.replset'].find_one()
        assert len(conf['members']) == 7
        assert sum(1 for memb in conf['members']
                   if 'arbiterOnly' in memb and memb['arbiterOnly']) == 1

        # check that 7 nodes are discovered
        assert len(self.tool.get_tagged('all')) == 7

    def test_large_replicaset_noarbiter(self):
        """mlaunch: start large replica set of 7 nodes without arbiter."""

        # start mongo process on free test port
        # (don't need journal for this test)
        self.run_tool("init --replicaset --nodes 7")

        # check if data directories exist
        assert os.path.exists(os.path.join(self.data_dir, 'replset'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs1'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs2'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs3'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs4'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs5'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs6'))
        assert os.path.exists(os.path.join(self.data_dir, 'replset/rs7'))

        # create mongo client for the next tests
        mc = MongoClient('localhost:%i' % self.port)

        # get rs.conf() and check for 7 members, no arbiters
        conf = mc['local']['system.replset'].find_one()
        assert len(conf['members']) == 7
        assert sum(1 for memb in conf['members']
                   if 'arbiterOnly' in memb and memb['arbiterOnly']) == 0

    def test_stop(self):
        """mlaunch: test stopping all nodes """

        self.run_tool("init --replicaset")
        self.run_tool("stop")

        # make sure all nodes are down
        nodes = self.tool.get_tagged('all')
        assert all(not self.tool.is_running(node) for node in nodes)

    def test_kill_default(self):
        """mlaunch: test killing all nodes with default signal."""

        # start sharded cluster and kill with default signal (15)
        self.run_tool("init --sharded 2 --single")
        self.run_tool("kill")

        # make sure all nodes are down
        nodes = self.tool.get_tagged('all')
        assert all(not self.tool.is_running(node) for node in nodes)

    def test_kill_sigterm(self):
        """mlaunch: test killing all nodes with SIGTERM."""

        # start nodes again, this time, kill with string "SIGTERM"
        self.run_tool("init --sharded 2 --single")
        self.run_tool("kill --signal SIGTERM")

        # make sure all nodes are down
        nodes = self.tool.get_tagged('all')
        assert all(not self.tool.is_running(node) for node in nodes)

    def test_kill_sigkill(self):
        """mlaunch: test killing all nodes with SIGKILL."""

        # start nodes again, this time, kill with signal 9 (SIGKILL)
        self.run_tool("init --sharded 2 --single")
        self.run_tool("kill --signal 9")

        # make sure all nodes are down
        nodes = self.tool.get_tagged('all')
        assert all(not self.tool.is_running(node) for node in nodes)

    def test_stop_start(self):
        """mlaunch: test stop and then re-starting nodes."""

        # start mongo process on free test port
        self.run_tool("init --replicaset")
        self.run_tool("stop")
        time.sleep(2)
        self.run_tool("start")

        # make sure all nodes are running
        nodes = self.tool.get_tagged('all')
        assert all(self.tool.is_running(node) for node in nodes)

    @unittest.skip('tags implementation not up to date')
    @pytest.mark.timeout(180)
    @pytest.mark.slow
    def test_kill_partial(self):
        """Test killing and restarting tagged groups on different tags."""

        # key is tag for command line, value is tag for get_tagged
        tags = ['shard01', 'shard 1', 'mongos', 'config 1', str(self.port)]

        # start large cluster
        self.run_tool("init --sharded 2 --replicaset --config 3 --mongos 3")

        # make sure all nodes are running
        nodes = self.tool.get_tagged('all')
        assert all(self.tool.is_running(node) for node in nodes)

        # go through all tags, stop nodes for each tag, confirm only
        # the tagged ones are down, start again
        for tag in tags:
            print("--------- %s" % tag)
            self.run_tool("kill %s" % tag)
            assert self.tool.get_tagged('down') == self.tool.get_tagged(tag)
            time.sleep(1)

            # short sleep, because travis seems to be sensitive and sometimes
            # fails otherwise
            self.run_tool("start")
            assert len(self.tool.get_tagged('down')) == 0
            time.sleep(1)

        # make sure primaries are running again
        # (we just failed them over above).
        # while True is ok, because test times out after some time
        while True:
            primaries = self.tool.get_tagged('primary')
            if len(primaries) == 2:
                break
            time.sleep(1)
            self.tool.discover()

        # test for primary, but as nodes lose their tags, needs to be manual
        self.run_tool("kill primary")
        assert len(self.tool.get_tagged('down')) == 2

    def test_restart_with_unkown_args(self):
        """mlaunch: test start command with extra unknown arguments."""

        # init environment (sharded, single shards ok)
        self.run_tool("init --single")

        # get verbosity of mongod, assert it is 0
        mc = MongoClient(port=self.port)
        loglevel = mc.admin.command(SON([('getParameter', 1), ('logLevel',
                                                               1)]))
        assert loglevel[u'logLevel'] == 0

        # stop and start nodes but pass in unknown_args
        self.run_tool("stop")

        # short sleep, because travis seems to be sensitive and
        # sometimes fails otherwise
        time.sleep(1)

        self.run_tool("start -vv")

        # compare that the nodes are restarted with the new unknown_args,
        # assert loglevel is now 2
        mc = MongoClient(port=self.port)
        loglevel = mc.admin.command(SON([('getParameter', 1), ('logLevel',
                                                               1)]))
        assert loglevel[u'logLevel'] == 2

        # stop and start nodes without unknown args again
        self.run_tool("stop")

        # short sleep, because travis seems to be sensitive and
        # sometimes fails otherwise
        time.sleep(1)

        self.run_tool("start")

        # compare that the nodes are restarted with the previous loglevel
        mc = MongoClient(port=self.port)
        loglevel = mc.admin.command(SON([('getParameter', 1), ('logLevel',
                                                               1)]))
        assert loglevel[u'logLevel'] == 0

    @unittest.skip('currently not a useful test')
    def test_start_stop_single_repeatedly(self):
        """Test starting and stopping single node in short succession."""

        # repeatedly start single node
        self.run_tool("init --single")

        for i in range(10):
            self.run_tool("stop")

            # short sleep, because travis seems to be sensitive and
            # sometimes fails otherwise
            time.sleep(1)

            self.run_tool("start")

    @pytest.mark.xfail(raises=SystemExit)
    def test_init_init_replicaset(self):
        """mlaunch: test calling init a second time on the replica set."""

        # init a replica set
        self.run_tool("init --replicaset")

        # now stop and init again, this should work if everything is
        # stopped and identical environment
        self.run_tool("stop")
        self.run_tool("init --replicaset")

        # but another init should fail with a SystemExit
        self.run_tool("init --replicaset")

    @unittest.skip('currently not a useful test')
    def test_start_stop_replicaset_repeatedly(self):
        """Test starting and stopping replica set in short succession."""

        # repeatedly start replicaset nodes
        self.run_tool("init --replicaset")

        for i in range(10):
            self.run_tool("stop")

            # short sleep, because travis seems to be sensitive and
            # sometimes fails otherwise
            time.sleep(1)

            self.run_tool("start")

    @pytest.mark.slow
    @pytest.mark.auth
    def test_repeat_all_with_auth(self):
        """Repeates all tests in this class (excluding itself) with auth."""
        tests = [t for t in inspect.getmembers(self,
                                               predicate=inspect.ismethod)
                 if t[0].startswith('test_')]

        self.use_auth = True

        for name, method in tests:
            # don't call any tests that use auth already (tagged with
            # 'auth' attribute), including this method
            if hasattr(method, 'auth'):
                continue

            setattr(method.__func__, 'description',
                    method.__doc__.strip() + ', with auth.')
            yield (method,)

        self.use_auth = False

    @pytest.mark.auth
    def test_replicaset_with_name(self):
        """mlaunch: test calling init on the replica set with given name."""

        self.run_tool("init --replicaset --name testrs")

        # create mongo client for the next tests
        mc = MongoClient('localhost:%i' % self.port)

        # get rs.conf() and check for its name
        conf = mc['local']['system.replset'].find_one()
        assert conf['_id'] == 'testrs'

    # TODO
    # - test functionality of --binarypath, --verbose

    # All tests that use auth need to be decorated with @pytest.mark.auth

    def helper_adding_default_user(self, environment):
        """Helper function for the next test: test_adding_default_user()."""

        self.run_tool("init %s --auth" % environment)

        # connect and authenticate with default credentials:
        # user / password on admin database
        mc = MongoClient('localhost:%i' % self.port)
        mc.admin.authenticate('user', password='password')

        # check if the user roles are correctly set to the default roles
        user = mc.admin.system.users.find_one()
        assert(set([x['role']
                    for x in user['roles']]) == set(self.tool.
                                                    _default_auth_roles))

    @pytest.mark.auth
    def test_adding_default_user(self):
        envs = (
            "--single",
            "--replicaset",
            "--sharded 2 --single",
            "--sharded 2 --replicaset",
            "--sharded 2 --single --config 3"
            )

        for env in envs:
            method = self.helper_adding_default_user
            setattr(method.__func__, 'description', method.__doc__.strip() +
                    ', with ' + env)
            yield (method, env)

    @pytest.mark.auth
    def test_adding_default_user_no_mongos(self):
        """mlaunch: test that even with --mongos 0 there is a user created."""

        self.run_tool("init --sharded 2 --single --mongos 0 --auth")

        # connect to config server instead to check for credentials (no mongos)
        ports = list(self.tool.get_tagged('config'))
        mc = MongoClient('localhost:%i' % ports[0])
        mc.admin.authenticate('user', password='password')

        # check if the user roles are correctly set to the default roles
        user = mc.admin.system.users.find_one()
        assert(set([x['role']
                    for x in user['roles']]) == set(self.tool.
                                                    _default_auth_roles))

    @pytest.mark.auth
    def test_adding_custom_user(self):
        """mlaunch: test custom username and password and custom roles."""

        self.run_tool("init --single --auth --username corben "
                      "--password fitzroy --auth-roles dbAdminAnyDatabase "
                      "readWriteAnyDatabase userAdminAnyDatabase")

        # connect and authenticate with default credentials:
        # user / password on admin database
        mc = MongoClient('localhost:%i' % self.port)
        mc.admin.authenticate('corben', password='fitzroy')

        # check if the user roles are correctly set to the specified roles
        user = mc.admin.system.users.find_one()
        print(user)
        assert(set([x['role']
                    for x in user['roles']]) == set(["dbAdminAnyDatabase",
                                                     "readWriteAnyDatabase",
                                                     "userAdminAnyDatabase"]))
        assert user['user'] == 'corben'

    def test_existing_environment(self):
        """mlaunch: test warning for overwriting an existing environment."""

        self.run_tool("init --single")
        self.run_tool("stop")
        try:
            self.run_tool("init --replicaset")
        except SystemExit as e:
            assert 'different environment already exists' in e.message

    @unittest.skip('mlaunch protocol upgrade is not needed at this point')
    def test_upgrade_v1_to_v2(self):
        """mlaunch: test upgrade from protocol version 1 to 2."""

        startup_options = {"name": "replset", "replicaset": True,
                           "dir": "./data", "authentication": False,
                           "single": False, "arbiter": False, "mongos": 1,
                           "binarypath": None, "sharded": None, "nodes": 3,
                           "config": 1, "port": 33333, "restart": False,
                           "verbose": False}

        # create directory
        self.run_tool("init --replicaset")
        self.run_tool("stop")

        # replace startup options
        with open(os.path.join(self.base_dir, 'test_upgrade_v1_to_v2',
                               '.mlaunch_startup'), 'w') as f:
            json.dump(startup_options, f, -1)

        # now start with old config and check if upgrade worked
        self.run_tool("start")
        with open(os.path.join(self.base_dir, 'test_upgrade_v1_to_v2',
                               '.mlaunch_startup'), 'rb') as f:
            startup_options = json.load(f)
            assert startup_options['protocol_version'] == 2

    def test_sharded_named_1(self):
        """mlaunch: test --sharded <name> for a single shard."""

        self.run_tool("init --sharded foo --single")
        assert len(self.tool.get_tagged('foo')) == 1

    def test_mlaunch_list(self):
        """mlaunch: test list command """

        self.run_tool("init --sharded 2 --replicaset --mongos 2")
        self.run_tool("list")

        # capture stdout and only keep from actual LIST output
        output = sys.stdout.getvalue().splitlines()
        output = output[output.index(next(o for o in output
                                          if o.startswith('PROCESS'))):]

        assert self.helper_output_has_line_with(['PROCESS', 'STATUS',
                                                 'PORT'], output) == 1
        assert self.helper_output_has_line_with(['mongos',
                                                 'running'], output) == 2
        assert self.helper_output_has_line_with(['config server',
                                                 'running'], output) == 1
        assert self.helper_output_has_line_with(['shard01'], output) == 1
        assert self.helper_output_has_line_with(['shard02'], output) == 1
        assert self.helper_output_has_line_with(['running',
                                                 'running'], output) == 9

    def helper_which(self, pgm):
        """equivalent of which command."""

        path = os.getenv('PATH')
        for p in path.split(os.path.pathsep):
            p = os.path.join(p, pgm)
            if os.path.exists(p) and os.access(p, os.X_OK):
                return p

    def test_mlaunch_binary_path_start(self):
        """Test if --binarypath is persistent between init and start."""

        # get true binary path (to test difference to not specifying one)
        path = self.helper_which('mongod')
        path = path[:path.rfind('/')]

        self.run_tool("init --single --binarypath %s" % path)
        self.run_tool("stop")

        self.run_tool("start")
        assert self.tool.loaded_args['binarypath'] == path
        assert self.tool.startup_info[str(self.port)].startswith('%s/mongod'
                                                                 % path)

        self.run_tool("stop")
        try:
            self.run_tool("start --binarypath /some/other/path")
            raise Exception
        except Exception:
            assert self.tool.args['binarypath'] == '/some/other/path'
            assert(self.tool.startup_info[str(self.port)].
                   startswith('/some/other/path/mongod'))

    @pytest.mark.xfail(raises=SystemExit)
    def test_single_and_arbiter(self):
        """mlaunch: test --single with --arbiter error."""

        self.run_tool("init --single --arbiter")

    def test_oplogsize_config(self):
        """mlaunch: test config server never receives --oplogSize parameter."""

        self.run_tool("init --sharded 1 --single --oplogSize 19 --verbose")
        output = sys.stdout.getvalue().splitlines()

        output_launch_config = next(o for o in output if '--configsvr' in o)
        assert '--oplogSize' not in output_launch_config


if __name__ == '__main__':

    # run individual tests with normal print output
    tml = TestMLaunch()
    tml.setup()
    tml.test_kill_partial()
    tml.teardown()



import argparse
import re
from datetime import timedelta
from operator import itemgetter

from mtools.mplotqueries.plottypes.base_type import BasePlotType

try:
    from matplotlib import __version__ as mpl_version
    import matplotlib.pyplot as plt
    from matplotlib.dates import date2num
    from matplotlib.patches import Polygon

except ImportError as error:
    raise ImportError("Can't import matplotlib. See "
                      "https://matplotlib.org/users/installing.html "
                      "for instructions on how to install matplotlib."
                      "Error: " + str(error))


class ScatterPlotType(BasePlotType):

    plot_type_str = 'scatter'
    sort_order = 3
    default_group_by = 'namespace'

    def __init__(self, args=None, unknown_args=None):
        BasePlotType.__init__(self, args, unknown_args)

        self.logscale = args['logscale']

        # parse arguments further to get --yaxis argument
        argparser = argparse.ArgumentParser("mplotqueries --type scatter")
        argparser.add_argument('--yaxis', '-y', action='store',
                               metavar='FIELD', default='duration')
        argparser.add_argument('--yaxis-regex-mode', action='store_true',
                               default=False)
        args = vars(argparser.parse_args(unknown_args))

        self.field = args['yaxis']
        self.regex_mode = args['yaxis_regex_mode']

        if self.regex_mode:
            self.ylabel = 'regex: "%s"' % self.field
        elif args['yaxis'] == 'duration':
            self.ylabel = 'duration in ms'
        else:
            self.ylabel = args['yaxis']

        self.durlines = []

    def accept_line(self, logevent):
        """Return True if the log line has the nominated yaxis field."""
        if self.regex_mode:
            return bool(re.search(self.field, logevent.line_str))
        else:
            return getattr(logevent, self.field) is not None

    def plot_group(self, group, idx, axis):
        # create x-coordinates for all log lines in this group
        x = date2num([logevent.datetime for logevent in self.groups[group]])

        color, marker = self.color_map(group)

        # duration plots require y coordinate and use plot_date
        if self.regex_mode:
            y = [float(re.search(self.field, logevent.line_str)
                       .group(1)) for logevent in self.groups[group]]
        else:
            y = [getattr(logevent, self.field)
                 for logevent in self.groups[group]]

        if self.logscale:
            axis.semilogy()

        group_label = group.replace('$', r'\$')
        artist = axis.plot_date(x, y, color=color, markeredgecolor='k',
                                marker=marker, alpha=0.8,
                                markersize=7, picker=5, label=group_label)[0]
        # add meta-data for picking
        artist._mt_plot_type = self
        artist._mt_group = group

        return artist

    def clicked(self, event):
        """
        Call if an element of this plottype is clicked.

        Implement in sub class.
        """
        group = event.artist._mt_group
        indices = event.ind

        # double click only supported on 1.2 or later
        major, minor, _ = mpl_version.split('.')
        if (int(major), int(minor)) < (1, 2) or not event.mouseevent.dblclick:
            for i in indices:
                print(self.groups[group][i].line_str)

        else:
            # toggle durline
            first = indices[0]
            logevent = self.groups[group][first]

            try:
                # remove triangle for this event
                idx = map(itemgetter(0), self.durlines).index(logevent)
                _, poly = self.durlines[idx]
                poly.remove()
                plt.gcf().canvas.draw()
                del self.durlines[idx]

            except ValueError:
                # construct triangle and add to list of durlines

                if self.args['optime_start']:
                    pts = [[date2num(logevent.datetime), 0],
                           [date2num(logevent.datetime), logevent.duration],
                           [date2num(logevent.datetime +
                                     timedelta(milliseconds=logevent.duration)
                                     ), 0]]
                else:
                    pts = [[date2num(logevent.datetime), 0],
                           [date2num(logevent.datetime), logevent.duration],
                           [date2num(logevent.datetime -
                                     timedelta(milliseconds=logevent.duration)
                                     ), 0]]

                poly = Polygon(pts, closed=True, alpha=0.2, linewidth=0,
                               facecolor=event.artist.get_markerfacecolor(),
                               edgecolor=None, zorder=-10000)

                ax = plt.gca()
                ax.add_patch(poly)
                plt.gcf().canvas.draw()

                self.durlines.append((logevent, poly))


class DurationLineType(ScatterPlotType):

    plot_type_str = 'durline'
    sort_order = 3
    default_group_by = 'namespace'

    def __init__(self, args=None, unknown_args=None):
        ScatterPlotType.__init__(self, args, unknown_args)
        self.args['optime_start'] = True

    def plot_group(self, group, idx, axis):
        # create x-coordinates for all log lines in this group
        x_start = date2num([logevent.datetime
                            for logevent in self.groups[group]])
        x_end = date2num([logevent.end_datetime
                          for logevent in self.groups[group]])

        color, marker = self.color_map(group)

        # duration plots require y coordinate and use plot_date
        y = [getattr(logevent, 'duration')
             for logevent in self.groups[group]]

        if self.logscale:
            axis.semilogy()

        # artist = axis.plot_date(x, y, color=color, markeredgecolor='k',
        #                         marker=marker, alpha=0.8, markersize=7,
        #                         picker=5, label=group)[0]

        artists = []
        labels = set()

        for i, (xs, xe, ye) in enumerate(zip(x_start, x_end, y)):
            artist = axis.plot_date([xs, xe], [0, ye], '-', color=color,
                                    alpha=0.8, linewidth=2,
                                    markersize=7, picker=5,
                                    label=None if group in labels else group
                                    )[0]

            labels.add(group)

            # add meta-data for picking
            artist._mt_plot_type = self
            artist._mt_group = group
            artist._mt_line_id = i
            artists.append(artist)

        return artists

    def clicked(self, event):
        group = event.artist._mt_group
        line_id = event.artist._mt_line_id
        print(self.groups[group][line_id].line_str)


class NScannedNPlotType(ScatterPlotType):

    plot_type_str = 'nscanned/n'
    default_group_by = 'namespace'

    def __init__(self, args=None, unknown_args=None):
        # Only call baseplot type constructor, we don't need argparser
        BasePlotType.__init__(self, args, unknown_args)

        self.ylabel = 'nscanned / n ratio'

    def accept_line(self, logevent):
        """Return True if the log line has a duration."""
        return getattr(logevent, 'nscanned') and getattr(logevent, 'nreturned')

    def plot_group(self, group, idx, axis):
        # create x-coordinates for all log lines in this group
        x = date2num([logevent.datetime for logevent in self.groups[group]])

        color, marker = self.color_map(group)

        y = [getattr(logevent, 'nscanned') /
             (float(getattr(logevent, 'nreturned'))
              if getattr(logevent, 'nreturned') != 0 else 1.0)
             for logevent in self.groups[group]]
        artist = axis.plot_date(x, y, color=color, marker=marker, alpha=0.8,
                                markersize=7, picker=5, label=group)[0]
        # add meta-data for picking
        artist._mt_plot_type = self
        artist._mt_group = group

        return artist


class DocsExaminedPlotType(ScatterPlotType):

    plot_type_str = 'docsExamined/n'
    default_group_by = 'namespace'

    def __init__(self, args=None, unknown_args=None):
        # Only call baseplot type constructor, we don't need argparser
        BasePlotType.__init__(self, args, unknown_args)

        self.ylabel = 'docsExamined / n ratio'

    def accept_line(self, logevent):
        """Return True if the log line has a duration."""
        # For backward compatibility the relevant logevent attribute is
        # currently nscannedObject.
        # This will have the value of nscannedObjects or equivalent
        # docsExamined metric.
        return getattr(logevent, 'nscannedObjects') and getattr(logevent,
                                                                'nreturned')

    def plot_group(self, group, idx, axis):
        # create x-coordinates for all log lines in this group
        x = date2num([logevent.datetime for logevent in self.groups[group]])

        color, marker = self.color_map(group)

        y = [getattr(logevent, 'nscannedObjects') /
             (float(getattr(logevent, 'nreturned'))
              if getattr(logevent, 'nreturned') != 0 else 1.0)
             for logevent in self.groups[group]]
        artist = axis.plot_date(x, y, color=color, marker=marker, alpha=0.8,
                                markersize=7, picker=5, label=group)[0]
        # add meta-data for picking
        artist._mt_plot_type = self
        artist._mt_group = group

        return artist



from datetime import MAXYEAR, MINYEAR, datetime

from dateutil.tz import tzutc

from mtools.util import OrderedDict
from mtools.util.grouping import Grouping

try:
    from matplotlib import cm
except ImportError as error:
    raise ImportError("Can't import matplotlib. See "
                      "https://matplotlib.org/users/installing.html "
                      "for instructions on how to install matplotlib."
                      "Error: " + str(error))


class BasePlotType(object):

    # 14 most distinguishable colors, according to
    # http://stackoverflow.com/questions/309149/generate-distinctly-different-rgb-colors-in-graphs
    colors = ['#0000FF', '#FF00F6', '#01FFFE', '#505050', '#909090',
              '#FF0000', '#00FF00', '#FFA6FE', '#FFDB66', '#006401',
              '#010067', '#95003A', '#007DB5', '#FFEEE8', '#774D00']

    color_index = 0
    markers = ['o', 's', '<', 'D']
    marker_index = 0

    sort_order = 0
    plot_type_str = 'base'
    default_group_by = None
    date_range = (datetime(MAXYEAR, 12, 31, tzinfo=tzutc()),
                  datetime(MINYEAR, 1, 1, tzinfo=tzutc()))

    def __init__(self, args=None, unknown_args=None):
        self.args = args
        self.unknown_args = unknown_args
        self.groups = OrderedDict()
        self.empty = True
        self.limits = None

        if self.args['optime_start']:
            self.xlabel = 'time (start of ops)'
        else:
            self.xlabel = 'time (end of ops)'

    def accept_line(self, logevent):
        """Return True if this PlotType can plot this line."""
        return True

    def add_line(self, logevent):
        """Append log line to this plot type."""
        key = None
        self.empty = False
        self.groups.setdefault(key, list()).append(logevent)

    @property
    def logevents(self):
        """Iterator yielding all logevents from groups dictionary."""
        for key in self.groups:
            for logevent in self.groups[key]:
                try:
                    yield logevent
                except StopIteration:
                    return

    @classmethod
    def color_map(cls, group):
        color = cls.colors[cls.color_index]
        cls.color_index += 1

        marker = cls.markers[cls.marker_index]
        if cls.color_index >= len(cls.colors):
            cls.marker_index += 1
            cls.marker_index %= len(cls.markers)
            cls.color_index %= cls.color_index

        return color, marker

    def group(self):
        """(re-)group all logevents by the given group."""
        if hasattr(self, 'group_by'):
            group_by = self.group_by
        else:
            group_by = self.default_group_by
            if self.args['group'] is not None:
                group_by = self.args['group']

        self.groups = Grouping(self.logevents, group_by)
        self.groups.move_items(None, 'others')
        self.groups.sort_by_size(group_limit=self.args['group_limit'],
                                 discard_others=self.args['no_others'])

    def plot_group(self, group, idx, axis):
        raise NotImplementedError("BasePlotType can't plot. "
                                  "Use a derived class instead")

    def clicked(self, event):
        """
        Call if an element of this plottype was clicked.

        Implement in sub class.
        """
        pass

    def plot(self, axis, ith_plot, total_plots, limits):
        self.limits = limits

        artists = []
        print(self.plot_type_str.upper() + " plot")
        print("%5s %9s  %s" % ("id", " #points", "group"))

        for idx, group in enumerate(self.groups):
            print("%5s %9s  %s" % (idx + 1, len(self.groups[group]), group))
            group_artists = self.plot_group(group, idx + ith_plot, axis)
            if isinstance(group_artists, list):
                artists.extend(group_artists)
            else:
                artists.append(group_artists)

        print()

        return artists



class BaseSection(object):
    """
    BaseSection class.

    All sections need to derive from it and add their arguments to the
    mloginfo.argparser object and determine if they are active.
    """

    filterArgs = []
    name = 'base'
    active = False

    def __init__(self, mloginfo):
        """Save command line arguments and set active to False by default."""
        # mloginfo object, use it to get access to argparser and other
        # class variables
        self.mloginfo = mloginfo

    def run(self):
        """Override this method in subclasses."""
        pass



from datetime import datetime, timedelta

from dateutil.tz import tzutc

from mtools.util.hci import DateTimeBoundaries


def test_dtb_within_boundaries_absolute():

    start = datetime(2012, 10, 14, tzinfo=tzutc())
    end = datetime(2013, 6, 2, tzinfo=tzutc())
    dtb = DateTimeBoundaries(start, end)
    from_dt, to_dt = dtb('Feb 18 2013', 'Feb 19 2013')
    assert from_dt == datetime(2013, 2, 18, tzinfo=tzutc())
    assert to_dt == datetime(2013, 2, 19, tzinfo=tzutc())


def test_dtb_from_before_start():

    start = datetime(2012, 10, 14, tzinfo=tzutc())
    end = datetime(2013, 6, 2, tzinfo=tzutc())
    dtb = DateTimeBoundaries(start, end)
    from_dt, to_dt = dtb('Sep 15 2012', 'Dec 1 2012')
    assert from_dt == start
    assert to_dt == datetime(2012, 12, 1, tzinfo=tzutc())


def test_dtb_to_after_end():

    start = datetime(2012, 10, 14, tzinfo=tzutc())
    end = datetime(2013, 6, 2, tzinfo=tzutc())
    dtb = DateTimeBoundaries(start, end)
    from_dt, to_dt = dtb('2013-01-15', '2016-03-02')
    assert from_dt == datetime(2013, 1, 15, tzinfo=tzutc())
    assert to_dt == end


def test_dtb_both_outside_bounds():

    start = datetime(2012, 10, 14, tzinfo=tzutc())
    end = datetime(2013, 6, 2, tzinfo=tzutc())
    dtb = DateTimeBoundaries(start, end)
    from_dt, to_dt = dtb('2000-01-01', '2050-12-31')
    assert from_dt == start
    assert to_dt == end


def test_dtb_keywords():

    start = datetime(2012, 10, 14, tzinfo=tzutc())
    end = datetime(2050, 1, 1, tzinfo=tzutc())
    dtb = DateTimeBoundaries(start, end)

    # start and end
    from_dt, to_dt = dtb('start', 'end')
    assert from_dt == start
    assert to_dt == end

    # today
    from_dt, to_dt = dtb('start', 'today')
    today = datetime.now().replace(hour=0, minute=0, second=0, microsecond=0,
                                   tzinfo=tzutc())
    assert to_dt == today

    # now, must be witin one second from now
    from_dt, to_dt = dtb('start', 'now')
    now = datetime.now().replace(tzinfo=tzutc())
    assert now - to_dt < timedelta(seconds=1)


def test_dtb_string2dt():

    start = datetime(2000, 1, 1, tzinfo=tzutc())
    end = datetime(2015, 6, 13, tzinfo=tzutc())
    dtb = DateTimeBoundaries(start, end)

    # without lower bound
    assert dtb.string2dt('') == start
    assert dtb.string2dt('2013') == datetime(2013, 1, 1, 0, 0,
                                             tzinfo=tzutc())
    assert dtb.string2dt('Aug 2011') == datetime(2011, 8, 1, 0, 0,
                                                 tzinfo=tzutc())
    assert dtb.string2dt('29 Sep 1978') == datetime(1978, 9, 29, 0, 0,
                                                    tzinfo=tzutc())

    # no year given, choose end year if still in log file, else year before
    assert dtb.string2dt('20 Mar') == datetime(2015, 3, 20, 0, 0,
                                               tzinfo=tzutc())
    assert dtb.string2dt('20 Aug') == datetime(2014, 8, 20, 0, 0,
                                               tzinfo=tzutc())

    # weekdays, always use last week of log file end
    assert dtb.string2dt('Sat') == datetime(2015, 6, 13, 0, 0, tzinfo=tzutc())
    assert dtb.string2dt('Wed') == datetime(2015, 6, 10, 0, 0, tzinfo=tzutc())
    assert dtb.string2dt('Sun') == datetime(2015, 6, 7, 0, 0, tzinfo=tzutc())

    # constants
    assert dtb.string2dt('start') == start
    assert dtb.string2dt('end') == end
    assert dtb.string2dt('today') == datetime.now().replace(hour=0, minute=0,
                                                            second=0,
                                                            microsecond=0,
                                                            tzinfo=tzutc())
    assert dtb.string2dt('yesterday') == (datetime.now()
                                          .replace(hour=0, minute=0, second=0,
                                                   microsecond=0,
                                                   tzinfo=tzutc()) -
                                          timedelta(days=1))

    # times
    assert dtb.string2dt('29 Sep 1978 13:06') == datetime(1978, 9, 29, 13, 6,
                                                          tzinfo=tzutc())
    assert dtb.string2dt('29 Sep 13:06') == datetime(2014, 9, 29, 13, 6,
                                                     tzinfo=tzutc())
    assert dtb.string2dt('13:06') == datetime(2000, 1, 1, 13, 6,
                                              tzinfo=tzutc())
    assert dtb.string2dt('13:06:15') == datetime(2000, 1, 1, 13, 6, 15,
                                                 tzinfo=tzutc())
    assert dtb.string2dt('13:06:15.214') == datetime(2000, 1, 1, 13, 6, 15,
                                                     214000, tzinfo=tzutc())
    assert dtb.string2dt('Wed 13:06:15') == datetime(2015, 6, 10, 13, 6, 15,
                                                     tzinfo=tzutc())

    # offsets
    assert dtb.string2dt('2013 +1d') == datetime(2013, 1, 2, 0, 0,
                                                 tzinfo=tzutc())
    assert dtb.string2dt('Sep 2011 +1mo') == datetime(2011, 10, 1, 0, 0,
                                                      tzinfo=tzutc())
    assert dtb.string2dt('29 Sep 1978 +3hours') == datetime(1978, 9, 29, 3, 0,
                                                            tzinfo=tzutc())
    assert dtb.string2dt('20 Mar +5min') == datetime(2015, 3, 20, 0, 5,
                                                     tzinfo=tzutc())
    assert dtb.string2dt('20 Aug -2day') == datetime(2014, 8, 18, 0, 0,
                                                     tzinfo=tzutc())
    assert dtb.string2dt('Sat -1d') == datetime(2015, 6, 12, 0, 0,
                                                tzinfo=tzutc())
    assert dtb.string2dt('Wed +4sec') == datetime(2015, 6, 10, 0, 0, 4,
                                                  tzinfo=tzutc())
    assert dtb.string2dt('Sun -26h') == datetime(2015, 6, 5, 22, 0,
                                                 tzinfo=tzutc())
    assert dtb.string2dt('29 Sep 1978 13:06 +59s') == datetime(1978, 9, 29,
                                                               13, 6, 59,
                                                               tzinfo=tzutc())
    assert dtb.string2dt('29 Sep 13:06 +120secs') == datetime(2014, 9, 29,
                                                              13, 8,
                                                              tzinfo=tzutc())
    # assert dtb.string2dt('13:06 -1week') == datetime(2014, 12, 25, 13, 6,
    #                                                  tzinfo=tzutc())
    # print dtb.string2dt('13:06:15 -16sec')
    # assert dtb.string2dt('13:06:15 -16sec') == datetime(2014, 1, 1, 13, 5,
    #                                                     59, tzinfo=tzutc())
    # assert dtb.string2dt('13:06:15.214 +1h') == datetime(2014, 1, 1, 14, 6,
    #                                                      15, 214000,
    #                                                      tzinfo=tzutc())
    assert dtb.string2dt('Wed 13:06:15 -1day') == datetime(2015, 6, 9, 13, 6,
                                                           15, tzinfo=tzutc())

    print(dtb.string2dt('start +3h'))
    assert dtb.string2dt('start +3h') == start + timedelta(hours=3)

    # offset only
    assert dtb.string2dt('-2d') == datetime(2015, 6, 11, tzinfo=tzutc())

    # test presence / absence of year and behavior for adjustment
    assert dtb.string2dt('July 30 2015') == datetime(2015, 7, 30,
                                                     tzinfo=tzutc())
    assert dtb.string2dt('July 30') == datetime(2014, 7, 30,
                                                tzinfo=tzutc())
    assert dtb.string2dt('1899 Nov 1') == datetime(1899, 11, 1, tzinfo=tzutc())

    # isoformat
    from_dt = datetime(2014, 8, 5, 20, 57, 7, tzinfo=tzutc())
    assert dtb.string2dt(from_dt.isoformat()) == datetime(2014, 8, 5, 20, 57,
                                                          7, tzinfo=tzutc())
    trydate = '2014-04-28T16:17:18.192Z'
    assert dtb.string2dt(trydate) == datetime(2014, 4, 28, 16, 17, 18, 192000,
                                              tzinfo=tzutc())

    # with lower_bounds
    lower = datetime(2013, 5, 2, 16, 21, 58, 123, tzinfo=tzutc())
    assert dtb.string2dt('', lower) == end
    assert dtb.string2dt('2013', lower) == datetime(2013, 1, 1, 0, 0,
                                                    tzinfo=tzutc())
    assert dtb.string2dt('Aug', lower) == datetime(2014, 8, 1, 0, 0,
                                                   tzinfo=tzutc())
    assert dtb.string2dt('+3sec', lower) == lower + timedelta(seconds=3)
    assert dtb.string2dt('+4min', lower) == lower + timedelta(minutes=4)
    assert dtb.string2dt('-5hours', lower) == lower - timedelta(hours=5)


if __name__ == '__main__':
    test_dtb_string2dt()



from collections import defaultdict

from .base_section import BaseSection
from mtools.util.log2code import Log2CodeConverter

try:
    from mtools.util.profile_collection import ProfileCollection
except ImportError:
    ProfileCollection = None


class DistinctSection(BaseSection):
    """
    DistinctSection class.

    This section shows a distinct view of all log lines matched with the
    Log2Code matcher. It will output sorted statistics of which logevent
    patterns where matched how often (most frequent first).
    """

    name = "distinct"
    log2code = Log2CodeConverter()

    def __init__(self, mloginfo):
        BaseSection.__init__(self, mloginfo)

        # add --restarts flag to argparser
        helptext = ('outputs distinct list of all log line by message '
                    'type (slow)')
        self.mloginfo.argparser_sectiongroup.add_argument('--distinct',
                                                          action='store_true',
                                                          help=helptext)

    @property
    def active(self):
        """Rreturn boolean if this section is active."""
        return self.mloginfo.args['distinct']

    def run(self):
        """Run each line through log2code and group by matched pattern."""
        if ProfileCollection and isinstance(self.mloginfo.logfile,
                                            ProfileCollection):
            print("\n    not available for system.profile collections\n")
            return

        codelines = defaultdict(lambda: 0)
        non_matches = 0

        # get log file information
        logfile = self.mloginfo.logfile
        if logfile.start and logfile.end and not self.mloginfo.args['verbose']:
            progress_start = self.mloginfo._datetime_to_epoch(logfile.start)
            progress_total = (self.mloginfo._datetime_to_epoch(logfile.end) -
                              progress_start)
        else:
            self.mloginfo.progress_bar_enabled = False

        for i, logevent in enumerate(self.mloginfo.logfile):
            cl, _ = self.log2code(logevent.line_str)

            # update progress bar every 1000 lines
            if self.mloginfo.progress_bar_enabled and (i % 1000 == 0):
                if logevent.datetime:
                    progress_curr = self.mloginfo._datetime_to_epoch(logevent
                                                                     .datetime)
                    (self.mloginfo
                     .update_progress(float(progress_curr - progress_start) /
                                      progress_total))

            if cl:
                codelines[cl.pattern] += 1
            else:
                if logevent.operation:
                    # skip operations (command, insert, update, delete,
                    # query, getmore)
                    continue
                if not logevent.thread:
                    # skip the lines that don't have a thread name
                    # (usually map/reduce or assertions)
                    continue
                if len(logevent.split_tokens) - logevent.datetime_nextpos <= 1:
                    # skip empty log messages (after thread name)
                    continue
                if ("warning: log line attempted" in logevent.line_str and
                        "over max size" in logevent.line_str):
                    # skip lines that are too long
                    continue

                # everything else is a real non-match
                non_matches += 1
                if self.mloginfo.args['verbose']:
                    print("couldn't match:" + logevent)

        # clear progress bar again
        if self.mloginfo.progress_bar_enabled:
            self.mloginfo.update_progress(1.0)

        if self.mloginfo.args['verbose']:
            print('')

        for cl in sorted(codelines, key=lambda x: codelines[x], reverse=True):
            print("%8i  %s" % (codelines[cl], " ... ".join(cl)))

        print('')
        if non_matches > 0:
            print("distinct couldn't match %i lines" % non_matches)
            if not self.mloginfo.args['verbose']:
                print("to show non-matched lines, run with --verbose.")



#!/usr/bin/env python3
"""Utility for working with datetimes."""

import re
from datetime import datetime, timedelta

from dateutil import parser
from dateutil.tz import tzutc
from dateutil.utils import default_tzinfo

from mtools.util import OrderedDict


class DateTimeBoundaries(object):
    """Object for working with datetime boundaries."""

    timeunits = ['secs', 'sec', 's', 'mins', 'min', 'm', 'months', 'month',
                 'mo', 'hours', 'hour', 'h', 'days', 'day', 'd', 'weeks',
                 'week', 'w', 'years', 'year', 'y']
    weekdays = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']

    dtRegexes = OrderedDict([
        # special constants
        ('constant', re.compile('(now|start|end|today|yesterday)' +
                                '($|\s+)')),
        # weekday: Mon, Wed, Sat
        ('weekday', re.compile('(' + '|'.join(weekdays) + ')' + '($|\s+)')),
        # 11:59:00.123, 1:13:12.004  (also match timezone postfix like Z or
        # +0700 or -05:30)
        # ('time',     re.compile('(?P<hour>\d{1,2}):(?P<minute>\d{2,2})' +
        # '(?::(?P<second>\d{2,2})(?:.(?P<microsecond>\d{3,3}))?)
        # ?(?P<timezone>[0-9Z:\+\-]+)?' + '($|\s+)')),
        # offsets: +3min, -20s, +7days  (see timeunits above)
        ('offset', re.compile('(?P<operator>[\+-])(?P<value>\d+)(?P<unit>' +
                              '|'.join(timeunits) + ')' + '($|\s+)'))
    ])

    def __init__(self, start, end):
        """Initialize DateTimeBoundaries with true start/end datetime objs."""
        if start > end:
            raise ValueError('Error in DateTimeBoundaries: end cannot '
                             'be before start datetime.')

        # make sure all datetimes are timezone-aware
        self.start = start
        if not self.start.tzinfo:
            self.start = self.start.replace(tzinfo=tzutc())

        self.end = end
        if not self.end.tzinfo:
            self.end = self.end.replace(tzinfo=tzutc())

    def string2dt(self, s, lower_bound=None):
        """Return datetime from a given string."""
        original_s = s

        result = {}
        dt = None

        # if s is completely empty, return start or end,
        # depending on what parameter is evaluated
        if s == '':
            return self.end if lower_bound else self.start

        # first try to match the defined regexes
        for idx in self.dtRegexes:
            regex = self.dtRegexes[idx]
            mo = regex.search(s)
            # if match was found, cut it out of original string and
            # store in result
            if mo:
                result[idx] = mo
                s = s[:mo.start(0)] + s[mo.end(0):]

        # handle constants
        if 'constant' in result:
            constant = result['constant'].group(0).strip()
            if constant == 'end':
                dt = self.end
            elif constant == 'start':
                dt = self.start
            elif constant == 'today':
                dt = datetime.now().replace(hour=0, minute=0, second=0,
                                            microsecond=0)
            elif constant == 'yesterday':
                dt = datetime.now().replace(hour=0, minute=0, second=0,
                                            microsecond=0) - timedelta(days=1)
            elif constant == 'now':
                dt = datetime.now()

        elif 'weekday' in result:
                weekday = result['weekday'].group(0).strip()
                # assume most-recently occured weekday in logfile
                most_recent_date = self.end.replace(hour=0, minute=0, second=0,
                                                    microsecond=0)
                offset = (most_recent_date.weekday() -
                          self.weekdays.index(weekday)) % 7
                dt = most_recent_date - timedelta(days=offset)

        # if anything remains unmatched, try parsing it with dateutil's parser
        if s.strip() != '':
            try:
                if dt:
                    dt = default_tzinfo(parser.parse(s, default=dt), tzutc)
                else:
                    # check if it's only time, then use the start dt as
                    # default, else just use the current year
                    if re.match('(?P<hour>\d{1,2}):(?P<minute>\d{2,2})'
                                '(?::(?P<second>\d{2,2})'
                                '(?:.(?P<microsecond>\d{3,3}))?)?'
                                '(?P<timezone>[0-9Z:\+\-]+)?$', s):
                        default = self.end if lower_bound else self.start
                    else:
                        default = datetime(self.end.year, 1, 1, 0, 0, 0)
                    default = default.replace(second=0, microsecond=0)

                    dt = parser.parse(s, default=default)

            except ValueError:
                raise ValueError("Error in DateTimeBoundaries: "
                                 "can't parse datetime from %s" % s)

        if not dt:
            dt = lower_bound or self.end

        # if no timezone specified, use the one from the logfile
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=self.start.tzinfo)

        # time is applied separately (not through the parser) so that string
        # containing only time don't use today as default date
        # (parser behavior)
        # if 'time' in result:
        #     dct = dict((k, int(v))
        #                for k,v in six.iteritems(result['time'].groupdict(0)))
        #     dct['microsecond'] *= 1000
        #     dt = dt.replace(**dct)

        # apply offset
        if 'offset' in result:

            # separate in operator, value, unit
            dct = result['offset'].groupdict()

            mult = 1
            if dct['unit'] in ['s', 'sec', 'secs']:
                dct['unit'] = 'seconds'
            elif dct['unit'] in ['m', 'min', 'mins']:
                dct['unit'] = 'minutes'
            elif dct['unit'] in ['h', 'hour', 'hours']:
                dct['unit'] = 'hours'
            elif dct['unit'] in ['d', 'day', 'days']:
                dct['unit'] = 'days'
            elif dct['unit'] in ['w', 'week', 'weeks']:
                dct['unit'] = 'days'
                mult = 7
            elif dct['unit'] in ['mo', 'month', 'months']:
                dct['unit'] = 'days'
                mult = 30.43
            elif dct['unit'] in ['y', 'year', 'years']:
                dct['unit'] = 'days'
                mult = 365.24

            if dct['operator'] == '-':
                mult *= -1

            dt = dt + eval('timedelta(%s=%i)' % (dct['unit'],
                                                 mult * int(dct['value'])))

        # if parsed datetime is out of bounds and no year specified,
        # try to adjust year
        year_present = re.search('\d{4,4}', original_s)

        if not year_present and 'constant' not in result:
            if (dt < self.start and
                    dt.replace(year=dt.year + 1) >= self.start and
                    dt.replace(year=dt.year + 1) <= self.end):
                dt = dt.replace(year=dt.year + 1)
            elif (dt > self.end and
                    dt.replace(year=dt.year - 1) >= self.start and
                    dt.replace(year=dt.year - 1) <= self.end):
                dt = dt.replace(year=dt.year - 1)

        return dt

    def __call__(self, from_str=None, to_str=None):
        """Set the boundaries based on `from` and `to` strings."""
        from_dt = self.string2dt(from_str, lower_bound=None)
        to_dt = self.string2dt(to_str, lower_bound=from_dt)

        if to_dt < from_dt:
            raise ValueError('Error in DateTimeBoundaries: lower bound is '
                             'greater than upper bound.')

        # limit from and to at the real boundaries
        if to_dt > self.end:
            to_dt = self.end

        if from_dt < self.start:
            from_dt = self.start

        return from_dt, to_dt


if __name__ == '__main__':

    dtb = DateTimeBoundaries(parser.parse('Apr 8 2014 13:00-0400'),
                             parser.parse('Apr 10 2014 16:21-0400'))
    lower, upper = dtb('2014-04-08T13:21-0400', '')
    print("lower %s" % lower)
    print("upper %s" % upper)
    print(dtb.string2dt("start +3h"))



#!/usr/bin/env python3

import json
import re
import sys
from datetime import datetime

import dateutil.parser
from dateutil.tz import tzutc

from mtools.util.pattern import json2pattern


class DateTimeEncoder(json.JSONEncoder):
    """Custom datetime encoder for json output."""

    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


class LogEvent(object):
    """
    Extract information from log line and store properties/variables.
    line_str: the original line string
    split_tokens: a list of string tokens after splitting line_str using
                  whitespace as split points
    datetime: a datetime object for the logevent. For logfiles created with
              version 2.4+, it also contains micro-seconds
    duration: the duration of a timed operation in ms
    thread: the thread name (e.g. "conn1234") as string
    operation: insert, update, remove, query, command, getmore, None
    namespace: the namespace of the operation, or None
    command: the type of command, if the operation was a "command"
    pattern: the query pattern for queries, updates, counts, etc
    ...
    Certain operations also add the number of affected/scanned documents.
    If applicable, the following variables are also set, otherwise the
    default is None: nscanned, ntoreturn, nreturned, ninserted, nupdated
    For performance reason, all fields are evaluated lazily upon first
    request.
    """

    # datetime handler for json encoding
    dthandler = lambda obj: obj.isoformat() if isinstance(obj,
                                                          datetime) else None

    weekdays = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
    months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep',
              'Oct', 'Nov', 'Dec']

    log_operations = ['query', 'insert', 'update', 'remove', 'getmore',
                      'command', 'aggregate', 'transaction']
    log_levels = ['D', 'F', 'E', 'W', 'I', 'U']
    log_components = ['-', 'ACCESS', 'COMMAND', 'CONTROL', 'GEO', 'INDEX',
                      'NETWORK', 'QUERY', 'REPL', 'SHARDING', 'STORAGE',
                      'JOURNAL', 'WRITE', 'TOTAL']

    def __init__(self, doc_or_str):
        self._debug = False
        self._year_rollover = False
        if isinstance(doc_or_str, bytes):
            doc_or_str = doc_or_str.decode("utf-8")

        if isinstance(doc_or_str, str) or (sys.version_info.major == 2 and
                                           isinstance(doc_or_str, unicode)):
            # create from string, remove line breaks at end of _line_str
            self.from_string = True
            self._line_str = doc_or_str.rstrip()
            self._profile_doc = None
            self._reset()
        else:
            self.from_string = False
            self._profile_doc = doc_or_str
            # docs don't need to be parsed lazily, they are fast
            self._parse_document()

    def _reset(self):
        self._debug = False
        self._split_tokens_calculated = False
        self._split_tokens = None

        self._duration_calculated = False
        self._duration = None

        self._datetime_calculated = False
        self._datetime = None
        self._datetime_nextpos = None
        self._datetime_format = None
        self._datetime_str = ''

        self._thread_calculated = False
        self._thread = None

        self._operation_calculated = False
        self._operation = None
        self._namespace = None

        self._pattern = None
        self._sort_pattern = None
        self._actual_query = None
        self._actual_sort = None

        # SERVER-36414 - parameters for slow transactions
        self._lsid = None
        self._txnNumber = None
        self._autocommit = None
        self._readConcern = None
        self._timeActiveMicros = None
        self._timeInactiveMicros = None
        self._readTimestamp = None
        self._terminationCause = None
        self._locks = None

        self._command_calculated = False
        self._command = None

        self._counters_calculated = False
        self._allowDiskUse = None

        self._bytesRead = None
        self._bytesWritten = None
        self._timeReadingMicros = None
        self._timeWritingMicros = None

        # TODO: refactor from the legacy names to modern
        # (eg: nscanned => keysExamined). Currently _extract_counters()
        # maps newer property names into legacy equivalents for
        # broader log file support.
        self._nscanned = None         # keysExamined
        self._nscannedObjects = None  # docsExamined
        self._ntoreturn = None
        self._nupdated = None         # nModified
        self._nreturned = None        # nReturned or nMatched (updates)
        self._ninserted = None        # nInserted
        self._ndeleted = None         # nDeleted

        self._numYields = None
        self._planSummary = None
        self._actualPlanSummary = None
        self._writeConflicts = None
        self._r = None
        self._w = None
        self._conn = None
        self._hostname = None

        self._level_calculated = False
        self._level = None
        self._component = None
        self.merge_marker_str = ''

        self._client_metadata_calculated = False
        self._client_metadata = None

    def set_line_str(self, line_str):
        """
        Set line_str.
        Line_str is only writeable if LogEvent was created from a string,
        not from a system.profile documents.
        """
        if not self.from_string:
            raise ValueError("can't set line_str for LogEvent created from "
                             "system.profile documents.")

        if line_str != self._line_str:
            self._line_str = line_str.rstrip()
            self._reset()

    def get_line_str(self):
        """Return line_str depending on source, logfile or system.profile."""
        if self.from_string:
            return ' '.join([s for s in [self.merge_marker_str,
                                         self._datetime_str,
                                         self._line_str] if s])
        else:
            return ' '.join([s for s in [self._datetime_str,
                                         self._line_str] if s])

    line_str = property(get_line_str, set_line_str)

    @property
    def split_tokens(self):
        """Split string into tokens (lazy)."""
        if not self._split_tokens_calculated:
            # split into items (whitespace split)
            self._split_tokens = self._line_str.split()
            self._split_tokens_calculated = True

        return self._split_tokens

    @property
    def duration(self):
        """Calculate duration if available (lazy)."""
        if not self._duration_calculated:
            self._duration_calculated = True

            # split_tokens = self.split_tokens
            line_str = self.line_str

            if (line_str
                and line_str.endswith('ms')
                and 'Scheduled new oplog query' not in line_str):

                try:
                    # find duration from end
                    space_pos = line_str.rfind(" ")
                    if space_pos == -1:
                        return
                    self._duration = int(line_str[line_str.rfind(" ") +
                                                  1:-2].replace(',', ''))
                except ValueError:
                    self._duration = None
            elif "flushing" in self.line_str:
                matchobj = re.search(r'flushing mmaps took (\d+)ms',
                                     self.line_str)
                if matchobj:
                    self._duration = int(matchobj.group(1))
            # SERVER-16176 - Logging of slow checkpoints
            elif "Checkpoint took" in self.line_str:
                matchobj = re.search("Checkpoint took ([\d]+) seconds to complete", self.line_str)
                if matchobj:
                    self._duration = int(matchobj.group(1)) * 1000

        return self._duration

    # SERVER-41349 - get hostname from the DNS log line
    @property
    def hostname(self):
        line_str = self.line_str
        groups = re.search("DNS resolution while connecting to ([\w.]+) took ([\d]+)ms", line_str)
        self._hostname = groups.group(1)
        return self._hostname


    @property
    def cursor(self):
        """Pull the cursor information if available (lazy)."""
        line_str = self.line_str
        # SERVER-28604 Checking reaped cursor information
        groups = re.search("Cursor id ([\w.]+) timed out, idle since ([^\n]*)", line_str)
        self._cursorid = groups.group(1)
        self._reapedtime = groups.group(2)

        return self._cursorid

    @property
    def datetime(self):
        """Extract datetime if available (lazy)."""
        if not self._datetime_calculated:
            self._datetime_calculated = True

            # if no datetime after 10 tokens, break to avoid parsing
            # very long lines
            split_tokens = self.split_tokens[:10]

            for offs in range(len(split_tokens)):
                dt = self._match_datetime_pattern(split_tokens[offs:offs + 4])
                if dt:
                    self._datetime = dt
                    self._datetime_nextpos = offs
                    if self._datetime_format.startswith("iso8601"):
                        self._datetime_nextpos += 1
                    else:
                        self._datetime_nextpos += 4

                    # separate datetime str and linestr
                    self._line_str = (' '.join(self.split_tokens
                                               [self._datetime_nextpos:]))

                    if self.level:
                        self._datetime_nextpos += 2

                    self._reformat_timestamp(self._datetime_format)
                    break

        return self._datetime

    @property
    def datetime_format(self):
        if not self._datetime_calculated:
            _ = self.datetime

        return self._datetime_format

    @property
    def datetime_nextpos(self):
        if self._datetime_nextpos is None and not self._datetime_calculated:
            _ = self.datetime
        return self._datetime_nextpos

    def set_datetime_hint(self, format, nextpos, rollover):
        self._datetime_format = format
        self._datetime_nextpos = nextpos
        self._year_rollover = rollover

        # Fast check if timestamp format changed.
        # If it has, trigger datetime evaluation.
        if format.startswith('ctime'):
            if (len(self.split_tokens) < 4 or
                    self.split_tokens[self._datetime_nextpos - 4] not in
                    self.weekdays):
                _ = self.datetime
                return False
            return True
        else:
            if len(self.split_tokens) == 0:
                # empty line, no need to parse datetime
                self._datetime_calculated = True
                return False
            try:
                if not (self.split_tokens[self._datetime_nextpos - 1][0]
                        .isdigit()):
                    # not the timestamp format that was hinted
                    _ = self.datetime
                    return False
            except Exception:
                pass
            return True

    def _match_datetime_pattern(self, tokens):
        """
        Match the datetime pattern at the beginning of the token list.
        There are several formats that this method needs to understand
        and distinguish between (see MongoDB's SERVER-7965):
        ctime-pre2.4    Wed Dec 31 19:00:00
        ctime           Wed Dec 31 19:00:00.000
        iso8601-utc     1970-01-01T00:00:00.000Z
        iso8601-local   1969-12-31T19:00:00.000+0500
        """
        # first check: less than 4 tokens can't be ctime
        assume_iso8601_format = len(tokens) < 4

        # check for ctime-pre-2.4 or ctime format
        if not assume_iso8601_format:
            weekday, month, day, time = tokens[:4]
            if (len(tokens) < 4 or (weekday not in self.weekdays) or
                    (month not in self.months) or not day.isdigit()):
                assume_iso8601_format = True

        if assume_iso8601_format:
            # sanity check, because the dateutil parser could interpret
            # any numbers as a valid date
            if not re.match(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{3}',
                            tokens[0]):
                return None

            # convinced that this is a ISO-8601 format, the dateutil parser
            # will do the rest
            dt = dateutil.parser.parse(tokens[0])
            self._datetime_format = "iso8601-utc" \
                if tokens[0].endswith('Z') else "iso8601-local"

        else:
            # assume current year unless self.year_rollover
            # is set (from LogFile)
            year = datetime.now().year
            dt = dateutil.parser.parse(' '.join(tokens[: 4]),
                                       default=datetime(year, 1, 1))

            if dt.tzinfo is None:
                dt = dt.replace(tzinfo=tzutc())

            if self._year_rollover and dt > self._year_rollover:
                dt = dt.replace(year=year - 1)

            self._datetime_format = "ctime" \
                if '.' in tokens[3] else "ctime-pre2.4"

        return dt

    @property
    def thread(self):
        """Extract thread name if available (lazy)."""
        if not self._thread_calculated:
            self._thread_calculated = True

            split_tokens = self.split_tokens

            if not self.datetime_nextpos:
                return None
            if len(split_tokens) <= self.datetime_nextpos:
                return None

            connection_token = split_tokens[self.datetime_nextpos]
            match = re.match(r'^\[([^\]]*)\]$', connection_token)
            if match:
                self._thread = match.group(1)

            if self._thread is not None:
                if self._thread in ['initandlisten', 'mongosMain']:
                    if len(split_tokens) >= 5 and split_tokens[-5][0] == '#':
                        self._conn = 'conn' + split_tokens[-5][1:]
                elif self._thread.startswith('conn'):
                    self._conn = self._thread
        return self._thread

    @property
    def conn(self):
        r"""
        Extract conn name if available (lazy).
        This value is None for all lines except the log lines related to
        connections, that is lines matching '\[conn[0-9]+\]' or
        '\[(initandlisten|mongosMain)\] .* connection accepted from'.
        """
        self.thread
        return self._conn

    @property
    def operation(self):
        """
        Extract operation if available (lazy).
        Operations: query, insert, update, remove, getmore, command
        """
        if not self._operation_calculated:
            self._operation_calculated = True
            self._extract_operation_and_namespace()

        return self._operation

    @property
    def namespace(self):
        """Extract namespace if available (lazy)."""
        if not self._operation_calculated:
            self._operation_calculated = True
            self._extract_operation_and_namespace()

        return self._namespace

    def _extract_operation_and_namespace(self):
        """
        Helper method to extract both operation and namespace from a logevent.
        It doesn't make sense to only extract one as they appear back to back
        in the token list.
        """
        split_tokens = self.split_tokens

        if not self._datetime_nextpos:
            # force evaluation of thread to get access to datetime_offset and
            # to protect from changes due to line truncation.
            _ = self.thread

        if not self._datetime_nextpos or (len(split_tokens) <=
                                          self._datetime_nextpos + 2):
            return

        op = split_tokens[self._datetime_nextpos + 1].lower()

        if op == 'warning:':
            # check if this log line got truncated
            if ("warning: log line attempted" in self._line_str and
                    "over max size" in self._line_str):
                self._datetime_nextpos = split_tokens.index('...')
                op = split_tokens[self._datetime_nextpos + 1]
            else:
                # unknown warning, bail out
                return

        if op in self.log_operations:
            self._operation = op
            self._namespace = split_tokens[self._datetime_nextpos + 2]

    @property
    def pattern(self):
        """Extract query pattern from operations."""
        if not self._pattern:

            # trigger evaluation of operation
            if (self.operation in ['query', 'getmore', 'update', 'remove'] or
                    self.command in ['count', 'findandmodify']):
                self._pattern = self._find_pattern('query: ')
                # Fallback check for q: variation (eg "remove" command in 3.6+)
                if self._pattern is None:
                    self._pattern = self._find_pattern('q: ')
            elif self.command == 'find':
                self._pattern = self._find_pattern('filter: ')

        return self._pattern

    @property
    def sort_pattern(self):
        """Extract query pattern from operations."""
        if not self._sort_pattern:

            # trigger evaluation of operation
            if self.operation in ['query', 'getmore']:
                self._sort_pattern = self._find_pattern('orderby: ')

        return self._sort_pattern

    @property
    def actual_query(self):
        """Extract the actual query (not pattern) from operations."""
        if not self._actual_query:

            # trigger evaluation of operation
            if (self.operation in ['query', 'getmore', 'update', 'remove'] or
                    self.command in ['count', 'findandmodify']):
                self._actual_query = self._find_pattern('query: ', actual=True)
            elif self.command == 'find':
                self._actual_query = self._find_pattern('filter: ',
                                                        actual=True)

        return self._actual_query

    @property
    def actual_sort(self):
        """Extract the actual sort key (not pattern) from operations."""
        if not self._actual_sort:

            # trigger evaluation of operation
            if self.operation in ['query', 'getmore']:
                self._actual_sort = self._find_pattern('orderby: ',
                                                        actual=True)

        return self._actual_sort

    @property
    def command(self):
        """Extract query pattern from operations."""
        if not self._command_calculated:

            self._command_calculated = True
            if self.operation == 'command':
                try:
                    command_idx = self.split_tokens.index('command:')
                    command = self.split_tokens[command_idx + 1]
                    if command == '{':
                        # workaround for <= 2.2 log files,
                        # where command was not listed separately
                        command = self.split_tokens[command_idx + 2][:-1]
                    self._command = command.lower()
                except ValueError:
                    pass

        return self._command

    @property
    def nscanned(self):
        """Extract nscanned or keysExamined counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._nscanned

    @property
    def timeActiveMicros(self):
        """Extract timeActiveMicros if available (lazy)."""

        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._timeActiveMicros

    @property
    def timeInactiveMicros(self):
        """Extract timeInactiveMicros if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._timeInactiveMicros

    @property
    def nscannedObjects(self):
        """
        Extract counters if available (lazy).
        Looks for nscannedObjects or docsExamined.
        """
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._nscannedObjects

    @property
    def ntoreturn(self):
        """Extract ntoreturn counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._ntoreturn

    @property
    def writeConflicts(self):
        """Extract ntoreturn counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._writeConflicts

    @property
    def nreturned(self):
        """
        Extract counters if available (lazy).
        Looks for nreturned, nReturned, or nMatched counter.
        """
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._nreturned

    @property
    def terminationCause(self):

        # Looks for terminationCause counter in Transaction logs.

        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._terminationCause

    @property
    def ninserted(self):
        """Extract ninserted or nInserted counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._ninserted

    @property
    def bytesRead(self):
        """Extract bytesRead counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._bytesRead

    @property
    def bytesWritten(self):
        """Extract bytesWritten counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._bytesWritten

    @property
    def timeReadingMicros(self):
        """Extract timeReadingMicros counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._timeReadingMicros

    @property
    def timeWritingMicros(self):
        """Extract timeWritingMicros counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._timeWritingMicros

    @property
    def ndeleted(self):
        """Extract ndeleted or nDeleted counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._ndeleted

    @property
    def allowDiskUse(self):
        """Extract allowDiskUse counter for aggregation if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._allowDiskUse

    @property
    def nupdated(self):
        """Extract nupdated or nModified counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._nupdated

    @property
    def numYields(self):
        """Extract numYields counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._numYields

    @property
    def readTimestamp(self):
        """Extract readTimeStamp counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._readTimestamp

    @property
    def planSummary(self):
        """Extract planSummary if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._planSummary

    @property
    def actualPlanSummary(self):
        """Extract planSummary including JSON if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._actualPlanSummary

    @property
    def r(self):
        """Extract read lock (r) counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._r

    @property
    def lsid(self):

        """Extract lsid counter if available (lazy)."""
        self._lsid = self._find_pattern('lsid: ', actual=True)
        return self._lsid

    @property
    def locks(self):
        """Extract locks counter for transactions if available (lazy)."""
        self._locks = self._find_pattern('locks:', actual=True)

        return self._locks

    @property
    def txnNumber(self):
        """Extract txnNumber counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._txnNumber

    @property
    def autocommit(self):

        """Extract autocommit counter for transactions if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._autocommit

    @property
    def readConcern(self):

        """Extract readConcern Level if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._readConcern

    @property
    def w(self):
        """Extract write lock (w) counter if available (lazy)."""
        if not self._counters_calculated:
            self._counters_calculated = True
            self._extract_counters()

        return self._w

    def _extract_counters(self):
        """Extract counters like nscanned and nreturned from the logevent."""
        # extract counters (if present)
        counters = ['nscanned', 'nscannedObjects', 'ntoreturn', 'nreturned',
                    'ninserted', 'nupdated', 'ndeleted', 'r', 'w', 'numYields',
                    'planSummary', 'writeConflicts', 'keyUpdates', 'bytesRead', 'bytesWritten', 'timeReadingMicros',
                    'timeWritingMicros', 'lsid', 'txnNumber', 'autocommit',  'allowDiskUse', 'level',
                    'timeActiveMicros', 'timeInactiveMicros', 'duration', 'readTimestamp', 'terminationCause']

        # TODO: refactor mtools to use current counter names throughout
        # Transitionary hack: mapping of current names into prior equivalents
        counter_equiv = {
            'datetime': 'datetime',
            'docsExamined': 'nscannedObjects',
            'keysExamined': 'nscanned',
            'nDeleted': 'ndeleted',
            'nInserted': 'ninserted',
            'nMatched': 'nreturned',
            'nModified': 'nupdated',
            'cursorid' : 'cursorid',
            'repaedtime' : 'reapedtime'
        }
        counters.extend(counter_equiv.keys())

        split_tokens = self.split_tokens

        # trigger operation evaluation to get access to offset
        if self.operation:
            for t, token in enumerate(split_tokens[self.datetime_nextpos +
                                                   2:]):
                for counter in counters:
                    if token.startswith('%s:' % counter):
                        try:
                            # Remap counter to standard name, if applicable
                            counter = counter_equiv.get(counter, counter)
                            if (counter == 'level' and token.startswith('level')):
                                    self._readConcern = (
                                    split_tokens[t + 1 + self.datetime_nextpos + 2].replace(',', ''))
                            elif (counter == 'readTimestamp' and token.startswith('readTimestamp')):
                                vars(self)['_' + counter] = (token.split(':')
                                [-1]).replace(',', '')
                            elif (counter == 'terminationCause' and token.startswith('terminationCause')):
                                vars(self)['_' + counter] = (token.split(':')
                                [-1]).replace(',', '')
                            else:
                                vars(self)['_' + counter] = int((token.split(':')
                                                             [-1]).replace(',',
                                                                           ''))

                            # extract allowDiskUse counter
                            if (counter == 'allowDiskUse' and token.startswith('allowDiskUse')):
                                # Splitting space between token and value
                                self._allowDiskUse = split_tokens[t + 1 + self.datetime_nextpos + 2].replace(',','')
                            else:
                                vars(self)['_' + counter] = int((token.split(':')[-1]).replace(',', ''))

                        except ValueError:
                            # see if this is a pre-2.5.2 numYields with space
                            # in between (e.g. "numYields: 2")
                            # https://jira.mongodb.org/browse/SERVER-10101
                            if (counter == 'numYields' and
                                    token.startswith('numYields')):
                                try:
                                    self._numYields = int((split_tokens[t + 1 + self.datetime_nextpos + 2]).replace(',', ''))
                                except ValueError:
                                    pass

                            if (counter == 'bytesRead' and
                                    token.startswith('bytesRead')):
                                try:
                                    self._bytesRead = int((split_tokens[t + 1 + self.datetime_nextpos + 2]).replace(',', ''))
                                except ValueError:
                                    pass
                            if (counter == 'bytesWritten' and
                                    token.startswith('bytesWritten')):
                                try:
                                    self._bytesWritten = int(
                                        (split_tokens[t + 1 + self.datetime_nextpos + 2]).replace(',', ''))
                                except ValueError:
                                    pass
                            if (counter == 'timeReadingMicros' and
                                    token.startswith('timeReadingMicros')):
                                try:
                                    self._timeReadingMicros = int(
                                        (split_tokens[t + 1 + self.datetime_nextpos + 2]).replace(',', ''))
                                except ValueError:
                                    pass
                            if (counter == 'timeWritingMicros' and
                                    token.startswith('timeWritingMicros')):
                                try:
                                    self._timeWritingMicros = int(
                                        (split_tokens[t + 1 + self.datetime_nextpos + 2]).replace(',', ''))
                                except ValueError:
                                    pass
                            if (counter == 'txnNumber' and
                                    token.startswith('txnNumber')):
                                    self._txnNumber = int((split_tokens[t + 1 + self.datetime_nextpos + 2]).replace(',', ''))
                            if (counter == 'autocommit' and
                                    token.startswith('autocommit')):
                                    self._autocommit = (split_tokens[t + 1 + self.datetime_nextpos + 2].replace(',', ''))
                            if (counter == 'lsid' and
                                    token.startswith('lsid')):
                                    self._lsid = (split_tokens[t + 2 + self.datetime_nextpos + 2].replace(',', ''))
                            if (counter == 'planSummary' and
                                    token.startswith('planSummary')):
                                try:
                                    self._planSummary = split_tokens[t + 1 + self.datetime_nextpos + 2]
                                    if self._planSummary:
                                        if split_tokens[t + 1 + self.datetime_nextpos + 3] != '{':
                                            self._actualPlanSummary = self._planSummary
                                        else:
                                            self._actualPlanSummary = '%s %s' % (
                                                self._planSummary,
                                                self._find_pattern('planSummary: %s' % self._planSummary, actual=True)
                                            )
                                except ValueError:
                                    pass

                        # token not parsable, skip
                        break

    @property
    def level(self):
        """Extract log level if available (lazy)."""
        if not self._level_calculated:
            self._level_calculated = True
            self._extract_level()
        return self._level

    @property
    def component(self):
        """Extract log component if available (lazy)."""
        self.level
        return self._component

    def _extract_level(self):
        """Extract level and component if available (lazy)."""
        if self._level is None:
            split_tokens = self.split_tokens

            if not split_tokens:
                self._level = False
                self._component = False
                return

            x = (self.log_levels.index(split_tokens[1])
                 if split_tokens[1] in self.log_levels else None)

            if x is not None:
                self._level = split_tokens[1]
                self._component = split_tokens[2]
            else:
                self._level = False
                self._component = False

    @property
    def client_metadata(self):
        """Return client metadata."""
        if not self._client_metadata_calculated:
            self._client_metadata_calculated = True

            line_str = self.line_str
            if (line_str and line_str.find('client metadata')):
                try:
                    metadata_pos = line_str.find("{")
                    if metadata_pos == -1:
                        return
                    else:
                        metadata = line_str[metadata_pos:]
                        # Make valid JSON by wrapping field names in quotes
                        metadata, _ = re.subn(r'([{,])\s*([^,{\s\'"]+)\s*:',
                                              ' \\1 "\\2" : ', metadata)

                        # Replace double-quoted platform values with single quote
                        platform = re.search(r'"platform"\s+:\s+"(.*)"', metadata)
                        if (platform):
                            platform = platform.group(1)
                            platform_esc, _ = re.subn(r'"', r"'", platform)
                            metadata, _ = re.subn(platform, platform_esc, metadata)

                        self._client_metadata = json.loads(metadata)
                except ValueError:
                    self._client_metadata = None

        return self._client_metadata

    def parse_all(self):
        """
        Trigger extraction of all information.
        These values are usually evaluated lazily.
        """
        tokens = self.split_tokens
        duration = self.duration
        datetime = self.datetime
        thread = self.thread
        operation = self.operation
        namespace = self.namespace
        pattern = self.pattern
        nscanned = self.nscanned
        nscannedObjects = self.nscannedObjects
        ntoreturn = self.ntoreturn
        nreturned = self.nreturned
        ninserted = self.ninserted
        ndeleted = self.ndeleted
        nupdated = self.nupdated
        numYields = self.numYields
        txnNumber = self.txnNumber
        w = self.w
        r = self.r

    def _find_pattern(self, trigger, actual=False):
        # get start of json query pattern
        start_idx = self.line_str.rfind(trigger)
        if start_idx == -1:
            # no query pattern found
            return None

        stop_idx = 0
        brace_counter = 0
        search_str = self.line_str[start_idx + len(trigger):]

        for match in re.finditer(r'{|}', search_str):
            stop_idx = match.start()
            if search_str[stop_idx] == '{':
                brace_counter += 1
            else:
                brace_counter -= 1
            if brace_counter == 0:
                break
        search_str = search_str[:stop_idx + 1].strip()
        if search_str:
            if actual:
                return search_str
            else:
                return json2pattern(search_str, debug=self._debug)
        else:
            return None

    def _reformat_timestamp(self, format, force=False):
        if format not in ['ctime', 'ctime-pre2.4', 'iso8601-utc',
                          'iso8601-local']:
            raise ValueError('invalid datetime format %s, choose from ctime, '
                             'ctime-pre2.4, iso8601-utc, iso8601-local.')

        if ((self.datetime_format is None or
                (self.datetime_format == format and
                    self._datetime_str != '')) and not force):
            return
        elif self.datetime is None:
            return
        elif format.startswith('ctime'):
            dt_string = (self.weekdays[self.datetime.weekday()] + ' ' +
                         self.datetime.strftime("%b %d %H:%M:%S"))
            # remove zero-padding from day number
            tokens = dt_string.split(' ')
            if tokens[2].startswith('0'):
                tokens[2] = tokens[2].replace('0', ' ', 1)
            dt_string = ' '.join(tokens)
            if format == 'ctime':
                dt_string += '.' + str(int(self.datetime.microsecond /
                                           1000)).zfill(3)
        elif format == 'iso8601-local':
            dt_string = self.datetime.isoformat()
            if self.datetime.utcoffset() is None:
                dt_string += '+00:00'
            ms_str = str(int(self.datetime.microsecond / 1000)).zfill(3)[:3]
            # change isoformat string to have 3 digit milliseconds and no :
            # in offset
            dt_string = re.sub(r'(\.\d+)?([+-])(\d\d):(\d\d)',
                               '.%s\\2\\3\\4' % ms_str, dt_string, count=1)
        elif format == 'iso8601-utc':
            if self.datetime.utcoffset():
                dt_string = self.datetime.astimezone(tzutc()).strftime("%Y-%m-"
                                                                       "%dT%H:"
                                                                       "%M:%S")
            else:
                dt_string = self.datetime.strftime("%Y-%m-%dT%H:%M:%S")
            dt_string += '.' + str(int(self.datetime.microsecond /
                                       1000)).zfill(3)[:3] + 'Z'

        # set new string and format
        self._datetime_str = dt_string
        self._datetime_format = format

    def __str__(self):
        """Default string conversion for LogEvent object is its line_str."""
        return str(self.line_str)

    def to_dict(self, labels=None):
        """Convert LogEvent object to a dictionary."""
        output = {}
        if labels is None:
            labels = ['line_str', 'split_tokens', 'datetime', 'operation',
                      'thread', 'namespace', 'nscanned', 'ntoreturn',
                      'nreturned', 'ninserted', 'nupdated', 'ndeleted',
                      'duration', 'r', 'w', 'numYields',  'cursorid', 'reapedtime',
                      'txtNumber', 'lsid', 'autocommit', 'readConcern',
                      'timeActiveMicros', 'timeInactiveMicros']

        for label in labels:
            value = getattr(self, label, None)
            if value is not None:
                output[label] = value

        return output

    def to_json(self, labels=None):
        """Convert LogEvent object to valid JSON."""
        output = self.to_dict(labels)
        return json.dumps(output, cls=DateTimeEncoder, ensure_ascii=False)

    def _parse_document(self):
        """Parse system.profile doc, copy all values to member variables."""
        self._reset()

        doc = self._profile_doc

        self._split_tokens_calculated = True
        self._split_tokens = None

        self._duration_calculated = True
        self._duration = doc[u'millis']

        self._datetime_calculated = True
        self._datetime = doc[u'ts']
        if self._datetime.tzinfo is None:
            self._datetime = self._datetime.replace(tzinfo=tzutc())
        self._datetime_format = None
        self._reformat_timestamp('ctime', force=True)

        self._thread_calculated = True
        self._thread = doc['thread']

        self._operation_calculated = True
        self._operation = doc[u'op']
        self._namespace = doc[u'ns']

        self._command_calculated = True
        if self.operation == 'command':
            self._command = doc[u'command'].keys()[0]

        # query pattern for system.profile events, all three cases.
        # See SERVER-13245
        if 'query' in doc:
            if 'query' in doc['query'] and isinstance(doc['query']['query'],
                                                      dict):
                self._pattern = str(doc['query']['query']).replace("'", '"')
            elif '$query' in doc['query']:
                self._pattern = str(doc['query']['$query']).replace("'", '"')
            else:
                self._pattern = str(doc['query']).replace("'", '"')

            # sort pattern
            if ('orderby' in doc['query'] and
                    isinstance(doc['query']['orderby'], dict)):
                self._sort_pattern = str(doc['query']
                                         ['orderby']).replace("'", '"')
            elif '$orderby' in doc['query']:
                self._sort_pattern = str(doc['query']
                                         ['$orderby']).replace("'", '"')
            else:
                self._sort_pattern = None

        self._counters_calculated = True
        self._nscanned = doc[u'nscanned'] if 'nscanned' in doc else None
        self._ntoreturn = doc[u'ntoreturn'] if 'ntoreturn' in doc else None
        self._nupdated = doc[u'nupdated'] if 'nupdated' in doc else None
        self._nreturned = doc[u'nreturned'] if 'nreturned' in doc else None
        self._ninserted = doc[u'ninserted'] if 'ninserted' in doc else None
        self._ndeleted = doc[u'ndeleted'] if 'ndeleted' in doc else None
        self._numYields = doc[u'numYield'] if 'numYield' in doc else None
        self._txnNumber = doc[u'txnNumber'] if 'txnNumber' in doc else None
        self._lsid = doc[u'lsid'] if 'lsid' in doc else None
        self._autocommit = doc[u'autocommit'] if 'autocommit' in doc else None
        self._readConcern = doc[u'level'] if 'level' in doc else None
        self._timeActiveMicros = doc[u'timeActiveMicros'] if 'timeActiveMicros' in doc else None
        self._timeInactiveMicros = doc[u'timeInactiveMicros'] if 'timeInactiveMicros' in doc else None
        self._duration = doc[u'duration'] if 'duration' in doc else None
        self._datetime = doc[u'datetime'] if 'datetime' in doc else None

        if u'lockStats' in doc:
            self._r = doc[u'lockStats'][u'timeLockedMicros'][u'r']
            self._w = doc[u'lockStats'][u'timeLockedMicros'][u'w']
            self._r_acquiring = doc[u'lockStats']['timeAcquiringMicros'][u'r']
            self._w_acquiring = doc[u'lockStats']['timeAcquiringMicros'][u'w']
            locks = 'w:%i' % self.w if self.w is not None else 'r:%i' % self.r
        elif u'locks' in doc:
            locks = json.dumps(doc[u'locks'])
        else:
            locks = ''

        # build a fake line_str
        payload = ''
        if 'query' in doc:
            payload += ('query: %s' % str(doc[u'query'])
                        .replace("u'", "'").replace("'", '"'))
        if 'command' in doc:
            payload += ('command: %s' % str(doc[u'command'])
                        .replace("u'", "'").replace("'", '"'))
        if 'updateobj' in doc:
            payload += (' update: %s' % str(doc[u'updateobj'])
                        .replace("u'", "'").replace("'", '"'))

        scanned = 'nscanned:%i' % self._nscanned if 'nscanned' in doc else ''
        yields = 'numYields:%i' % self._numYields if 'numYield' in doc else ''
        duration = '%ims' % self.duration if self.duration is not None else ''

        self._line_str = (f'''[{self.thread}] {self.operation} {self.namespace} {payload} '''
                          f'''{scanned} {yields} locks(micros) {locks} '''
                          f'''{duration}''')



#!/usr/bin/env python3

import os
import re
import sys
from datetime import datetime
from math import ceil

from mtools.util.input_source import InputSource
from mtools.util.logevent import LogEvent


class LogFile(InputSource):
    """Log file wrapper class. Handles open file streams or stdin."""

    def __init__(self, filehandle):
        """Provide logfile as open file stream or stdin."""
        self.filehandle = filehandle
        self.name = filehandle.name

        self.from_stdin = filehandle.name == "<stdin>"
        self._bounds_calculated = False
        self._start = None
        self._end = None
        self._filesize = None
        self._num_lines = None
        self._restarts = None
        self._binary = None
        self._timezone = None
        self._hostname = None
        self._port = None
        self._rs_state = None

        self._repl_set = None
        self._repl_set_members = None
        self._repl_set_version = None
        self._repl_set_protocol = None

        self._storage_engine = None

        self._datetime_format = None
        self._year_rollover = None

        self._shards = None
        self._csrs = None

        self._chunks_moved_from = None
        self._chunks_moved_to = None
        self._chunk_splits = None

        # Track previous file position for loop detection in _find_curr_line()
        self.prev_pos = None

        self._has_level = None

        # make sure bounds are calculated before starting to iterate,
        # including potential year rollovers
        self._calculate_bounds()

    def __getstate__(self):
        """
        Return state values to be pickled.
        """
        if (self.name != '<stdin>'):
            return (self.name)
        else:
            return ()

    def __setstate__(self, state):
        """
        Restore state from the unpickled state values.
        """
        self.name = state

    @property
    def start(self):
        """
        Lazy evaluation of start and end of logfile.

        Returns None for stdin input currently.
        """
        if not self._start:
            self._calculate_bounds()
        return self._start

    @property
    def end(self):
        """
        Lazy evaluation of start and end of logfile.

        Returns None for stdin input currently.
        """
        if not self._end:
            self._calculate_bounds()
        return self._end

    @property
    def timezone(self):
        """Lazy evaluation of timezone of logfile."""
        if not self._timezone:
            self._calculate_bounds()
        return self._timezone

    @property
    def filesize(self):
        """
        Lazy evaluation of start and end of logfile.

        Returns None for stdin input currently.
        """
        if self.from_stdin:
            return None
        if not self._filesize:
            self._calculate_bounds()
        return self._filesize

    @property
    def datetime_format(self):
        """Lazy evaluation of the datetime format."""
        if not self._datetime_format:
            self._calculate_bounds()
        return self._datetime_format

    @property
    def has_level(self):
        """Lazy evaluation of the whether the logfile has any level lines."""
        if self._has_level is None:
            self._iterate_lines()
        return self._has_level

    @property
    def year_rollover(self):
        """Lazy evaluation of the datetime format."""
        if self._year_rollover is None:
            self._calculate_bounds()
        return self._year_rollover

    @property
    def num_lines(self):
        """
        Lazy evaluation of the number of lines.

        Returns None for stdin input currently.
        """
        if self.from_stdin:
            return None
        if not self._num_lines:
            self._iterate_lines()
        return self._num_lines

    @property
    def restarts(self):
        """Lazy evaluation of all restarts."""
        if not self._num_lines:
            self._iterate_lines()
        return self._restarts

    @property
    def rs_state(self):
        """Lazy evaluation of all restarts."""
        if not self._num_lines:
            self._iterate_lines()
        return self._rs_state

    @property
    def binary(self):
        """Lazy evaluation of the binary name."""
        if not self._num_lines:
            self._iterate_lines()
        return self._binary

    @property
    def hostname(self):
        """Lazy evaluation of the binary name."""
        if not self._num_lines:
            self._iterate_lines()
        return self._hostname

    @property
    def port(self):
        """Lazy evaluation of the binary name."""
        if not self._num_lines:
            self._iterate_lines()
        return self._port

    @property
    def versions(self):
        """Return all version changes."""
        versions = []
        for v, _ in self.restarts:
            if len(versions) == 0 or v != versions[-1]:
                versions.append(v)
        return versions

    @property
    def repl_set(self):
        """Return the replSet (if available)."""
        if not self._num_lines:
            self._iterate_lines()
        return self._repl_set

    @property
    def repl_set_members(self):
        """Return the replSet (if available)."""
        if not self._num_lines:
            self._iterate_lines()
        return self._repl_set_members

    @property
    def repl_set_version(self):
        """Return the replSet (if available)."""
        if not self._num_lines:
            self._iterate_lines()
        return self._repl_set_version

    @property
    def repl_set_protocol(self):
        """Return the replSet protocolVersion (if available)."""
        if not self._num_lines:
            self._iterate_lines()
        return self._repl_set_protocol

    @property
    def storage_engine(self):
        """Return storage engine if available."""
        if not self._num_lines:
            self._iterate_lines()
        return self._storage_engine

    @property
    def shards(self):
        """Lazily return the shards (if available)"""
        if not self._shards:
            self._find_sharding_info()
        return self._shards

    @property
    def csrs(self):
        """Lazily return the CSRS (if available)"""
        if not self._csrs:
            self._find_sharding_info()
        return self._csrs

    @property
    def chunks_moved_to(self):
        """Lazily return the chunks moved to this shard (if available)"""
        if not self._chunks_moved_to:
            self._find_sharding_info()
        return self._chunks_moved_to

    @property
    def chunks_moved_from(self):
        """Lazily return the chunks moved from this shard (if available)"""
        if not self._chunks_moved_from:
            self._find_sharding_info()
        return self._chunks_moved_from

    @property
    def chunk_splits(self):
        """Lazily return the chunks split in this shard (if available)"""
        if not self._chunk_splits:
            self._find_sharding_info()
        return self._chunk_splits


    def next(self):
        """Get next line, adjust for year rollover and hint datetime format."""
        # use readline here because next() iterator uses internal readahead
        # buffer so seek position is wrong
        line = self.filehandle.readline()

        if isinstance(line, bytes):
            line = line.decode('utf-8', 'replace')

        if line == '':
            raise StopIteration
        line = line.rstrip('\n')
        le = LogEvent(line)

        # hint format and nextpos from previous line
        if self._datetime_format and self._datetime_nextpos is not None:
            ret = le.set_datetime_hint(self._datetime_format,
                                       self._datetime_nextpos,
                                       self.year_rollover)
            if not ret:
                # logevent indicates timestamp format has changed,
                # invalidate hint info
                self._datetime_format = None
                self._datetime_nextpos = None
        elif le.datetime:
            # gather new hint info from another logevent
            self._datetime_format = le.datetime_format
            self._datetime_nextpos = le._datetime_nextpos

        return le

    def __iter__(self):
        """
        Iterate over LogFile object.

        Return a LogEvent object for each line (generator).
        """
        le = None

        while True:
            try:
                le = self.next()
            except StopIteration as e:
                # end of log file, get end date
                if not self.end and self.from_stdin:
                    if le and le.datetime:
                        self._end = le.datetime

                # future iterations start from the beginning
                if not self.from_stdin:
                    self.filehandle.seek(0)

                # return (instead of raising StopIteration exception) per PEP 479
                return

            # get start date for stdin input
            if not self.start and self.from_stdin:
                if le and le.datetime:
                    self._start = le.datetime

            try:
                yield le
            except StopIteration:
                return

    states = (['PRIMARY', 'SECONDARY', 'DOWN', 'STARTUP', 'STARTUP2',
               'RECOVERING', 'ROLLBACK', 'ARBITER', 'UNKNOWN'])

    def __len__(self):
        """Return the number of lines in a log file."""
        return self.num_lines

    def _iterate_lines(self):
        """Count number of lines (can be expensive)."""
        self._num_lines = 0
        self._restarts = []
        self._rs_state = []
        
        ln = 0
        for ln, line in enumerate(self.filehandle):
            if isinstance(line, bytes):
                line = line.decode("utf-8", "replace")

            if (self._has_level is None and
                    line[28:31].strip() in LogEvent.log_levels and
                    line[31:39].strip() in LogEvent.log_components):
                self._has_level = True

            # find version string (fast check to eliminate most lines)
            if "version" in line[:100]:
                logevent = LogEvent(line)
                restart = self._check_for_restart(logevent)
                if restart:
                    self._restarts.append((restart, logevent))

            # look for hostname and port which will be logged with pid
            if not self._hostname:
                if " pid=" in line:
                    match = re.search('port=(?P<port>\d+).*host=(?P<host>\S+)',
                                    line)
                    if match:
                        self._hostname = match.group('host')
                        self._port = match.group('port')

            """ For 3.0 the "[initandlisten] options:" long entry contained the
                "engine" field if WiredTiger was the storage engine. There were
                only two engines, MMAPv1 and WiredTiger
            """
            if "[initandlisten] options:" in line:
                match = re.search('replSet: "(?P<replSet>\S+)"', line)
                if match:
                    self._repl_set = match.group('replSet')

                match = re.search('engine: "(?P<engine>\S+)"', line)
                if match:
                    self._storage_engine = match.group('engine')
                else:
                    self._storage_engine = 'mmapv1'

            """ For 3.2 the "[initandlisten] options:" no longer contains the
                "engine" field So now we have to look for the "[initandlisten]
                wiredtiger_open config:" which was present in 3.0, but would
                now tell us definitively that wiredTiger is being used
            """
            if "[initandlisten] wiredtiger_open config:" in line:
                self._storage_engine = 'wiredTiger'

            if "command admin.$cmd command: { replSetInitiate:" in line:
                match = re.search('{ _id: "(?P<replSet>\S+)", '
                                  'members: (?P<replSetMembers>[^]]+ ])', line)
                if match:
                    self._repl_set = match.group('replSet')
                    self._repl_set_members = match.group('replSetMembers')

            # Replica set config logging in MongoDB 3.0+
            new_config = ("New replica set config in use: ")
            if new_config in line:
                match = re.search('{ _id: "(?P<replSet>\S+)", '
                                  'version: (?P<replSetVersion>\d+), ', line)
                if match:
                    self._repl_set = match.group('replSet')
                    self._repl_set_version = match.group('replSetVersion')
                match = re.search(', protocolVersion: (?P<replSetProtocol>\d+), ', line)
                if match:
                    self._repl_set_protocol = match.group('replSetProtocol')
                match = re.search('members: (?P<replSetMembers>[^]]+ ])', line)
                if match:
                    self._repl_set_members = match.group('replSetMembers')


            # if ("is now in state" in line and
            #        next(state for state in states if line.endswith(state))):
            if "is now in state" in line:
                tokens = line.split()
                # 2.6
                if tokens[1].endswith(']'):
                    pos = 4
                else:
                    pos = 5
                host = tokens[pos]
                rs_state = tokens[-1]
                state = (host, rs_state, LogEvent(line))
                self._rs_state.append(state)
                continue

            if "[rsMgr] replSet" in line:
                tokens = line.split()
                if self._hostname:
                    host = self._hostname + ':' + self._port
                else:
                    host = os.path.basename(self.name)
                host += ' (self)'
                if tokens[-1] in self.states:
                    rs_state = tokens[-1]
                else:
                    # 2.6
                    if tokens[1].endswith(']'):
                        pos = 2
                    else:
                        pos = 6
                    rs_state = ' '.join(tokens[pos:])

                state = (host, rs_state, LogEvent(line))
                self._rs_state.append(state)
                continue

        self._num_lines = ln + 1

        # reset logfile
        self.filehandle.seek(0)

    def _check_for_restart(self, logevent):

        if (logevent.thread == 'initandlisten' and
                "db version v" in logevent.line_str):
            self._binary = 'mongod'

        elif logevent.thread == 'mongosMain' and ('MongoS' in logevent.line_str or 
            'mongos' in logevent.line_str):
            self._binary = 'mongos'

        else:
            return False

        version = re.search(r'(?:db|mongos) version v(\d\.\d\.\d+)', logevent.line_str)

        if version:
            version = version.group(1)
            return version
        else:
            return False

    def _calculate_bounds(self):
        """Calculate beginning and end of logfile."""
        if self._bounds_calculated:
            # Assume no need to recalc bounds for lifetime of a Logfile object
            return

        if self.from_stdin:
            return False

        # we should be able to find a valid log line within max_start_lines
        max_start_lines = 10
        lines_checked = 0

        # get start datetime
        for line in self.filehandle:
            logevent = LogEvent(line)
            lines_checked += 1
            if logevent.datetime:
                self._start = logevent.datetime
                self._timezone = logevent.datetime.tzinfo
                self._datetime_format = logevent.datetime_format
                self._datetime_nextpos = logevent._datetime_nextpos
                break
            if lines_checked > max_start_lines:
                break

        # sanity check before attempting to find end date
        if (self._start is None):
            raise SystemExit("Error: <%s> does not appear to be a supported "
                             "MongoDB log file format" % self.filehandle.name)

        # get end datetime (lines are at most 10k,
        # go back 30k at most to make sure we catch one)
        self.filehandle.seek(0, 2)
        self._filesize = self.filehandle.tell()
        self.filehandle.seek(-min(self._filesize, 30000), 2)

        for line in reversed(self.filehandle.readlines()):
            logevent = LogEvent(line)
            if logevent.datetime:
                self._end = logevent.datetime
                break

        # if there was a roll-over, subtract 1 year from start time
        if self._end < self._start:
            self._start = self._start.replace(year=self._start.year - 1)
            self._year_rollover = self._end
        else:
            self._year_rollover = False

        # reset logfile
        self.filehandle.seek(0)
        self._bounds_calculated = True

        return True

    def _find_curr_line(self, prev=False):
        """
        Internal helper function.

        Find the current (or previous if prev=True) line in a log file based on
        the current seek position.
        """
        curr_pos = self.filehandle.tell()

        # jump back 15k characters (at most) and find last newline char
        jump_back = min(self.filehandle.tell(), 15000)
        self.filehandle.seek(-jump_back, 1)
        buff = self.filehandle.read(jump_back)
        self.filehandle.seek(curr_pos, 0)

        if prev and self.prev_pos is not None and self.prev_pos == curr_pos:
            # Number of characters to show before/after the log offset
            error_context = 300
            self.filehandle.seek(-error_context, 1)
            buff = self.filehandle.read(curr_pos)
            hr = "-" * 60
            print("Fatal log parsing loop detected trying to find previous "
                  "log line near offset %s in %s:\n\n%s\n%s\n"
                  "<--- (current log parsing offset) \n%s\n%s\n"
                  % (curr_pos, self.name, hr, buff[:error_context],
                     buff[error_context:error_context + 1], hr),
                  file=sys.stderr)
            raise SystemExit("Cannot parse %s with requested options"
                             % self.filehandle.name)
        else:
            self.prev_pos = curr_pos

        if isinstance(buff, bytes):
            buff = buff.decode("utf-8", "replace")

        newline_pos = buff.rfind('\n')
        if prev:
            newline_pos = buff[:newline_pos].rfind('\n')

        # move back to last newline char
        if newline_pos == -1:
            self.filehandle.seek(0)
            return self.next()

        self.filehandle.seek(newline_pos - jump_back + 1, 1)

        # roll forward until we found a line with a datetime
        try:
            logevent = self.next()
            while not logevent.datetime:
                logevent = self.next()

            return logevent
        except StopIteration:
            # reached end of file
            return None

    def _find_sharding_info(self):
        """
        Iterate over file and find any sharding related information
        """
        self._shards = []
        self._chunks_moved_from = []
        self._chunks_moved_to = []
        self._chunk_splits = []
        
        prev_line = ""

        for line in self.filehandle:
            if isinstance(line, bytes):
                line = line.decode("utf-8", "replace")

            if self.binary == "mongos":
        
                if "Starting new replica set monitor for" in line:
                    if "[mongosMain]" in line:
                        match = re.search("for (?P<csrsName>\w+)/"
                                          "(?P<replSetMembers>\S+)", line)
                        if match:
                            csrs_info = (match.group('csrsName'),
                                         match.group('replSetMembers'))
                            self._csrs = csrs_info
                    else:
                        match = re.search("for (?P<shardName>\w+)/"
                                          "(?P<replSetMembers>\S+)", line)
                        if match:
                            shard_info = (match.group('shardName'),
                                        match.group('replSetMembers'))
                            self._shards.append(shard_info)

            elif self.binary == "mongod":
                logevent = LogEvent(line)
                if "New replica set config in use" in line:
                    
                    if "configsvr: true" in line:
                        match = re.search(' _id: "(?P<replSet>\S+)".*'
                                          'members: (?P<replSetMembers>[^]]+ ])', line)
                        if match:
                            self._csrs = (
                                match.group('replSet'),
                                match.group('replSetMembers')
                            )

                if "Starting new replica set monitor for" in line:
                    match = re.search("for (?P<replSet>\w+)/"
                                        "(?P<replSetMembers>\S+)", line)
                    if match:
                        if self._csrs and match.group('replSet') != self._csrs[0]:
                            self._shards.append((
                                match.group('replSet'),
                                match.group('replSetMembers')
                            ))
                        elif not self._csrs:
                            self._csrs = (
                                match.group('replSet'),
                                match.group('replSetMembers')
                            )
    
            if "moveChunk.from" in line:
                logevent = LogEvent(line)
                match = re.search('ns: "(?P<namespace>\S+)".*'
                                  'details: { (?P<range>.*\}).*'
                                  'to: "(?P<movedTo>\S+)".*note: "(?P<note>\S+)"', line)
                if match:
                    time = logevent.datetime
                    chunk_range = match.group('range')
                    namespace = match.group('namespace')
                    moved_to = match.group('movedTo')
                    note = match.group('note')
                    
                    if note == "success":
                        errmsg = None
                        steps = re.findall('(?P<steps>step \d of \d): (?P<stepTimes>\d+)', line)
                    else:
                        match = re.search(':: caused by :: (?P<errmsg>\S+):', prev_line)
                        steps = None
                        if match:
                            errmsg = match.group('errmsg')
                        else:
                            errmsg = "Unknown"

                    chunk_migration = (time, chunk_range, moved_to, namespace, steps, note, errmsg)

                    self._chunks_moved_from.append(chunk_migration)

            if "moveChunk.to" in line:
                logevent = LogEvent(line)
                match = re.search('ns: "(?P<namespace>\S+)".*'
                                  'details: { (?P<range>.*\}).*.*note: "(?P<note>\S+)"', line)
                if match:
                    time = logevent.datetime
                    chunk_range = match.group('range')
                    namespace = match.group('namespace')
                    # TODO: alter this to find moved from shard name when SERVER-45770 TICKET is added
                    moved_from = "Unknown"
                    note = match.group('note')

                    if note == "success":
                        errmsg = None
                        steps = re.findall('(?P<steps>step \d of \d): (?P<stepTimes>\d+)', line)
                    else:
                        steps = None
                        match = re.search('errmsg: "(?P<errmsg>.*)"', line)
                        if match:
                            errmsg = match.group('errmsg')

                    chunk_migration = (time, chunk_range, moved_from, namespace, steps, note, errmsg)

                    self._chunks_moved_to.append(chunk_migration)

            if "Finding the split vector for" in line:
                logevent = LogEvent(line)
                match = re.search('for (?P<namespace>\S+).*'
                                  'numSplits: (?P<numSplits>\d+)', line)
                if match:
                    time = logevent.datetime
                    split_range = None
                    namespace = match.group("namespace")
                    numSplits = match.group('numSplits')
                    success = None
                    time_taken = 0
                    error = None
                    self._chunk_splits.append((time, split_range, namespace, numSplits, success, time_taken, error))
            elif "splitVector" in line:
                logevent = LogEvent(line)
                match = re.search('splitVector: "(?P<namespace>\S+)".*,'
                                  ' (?P<range>min:.*), max.*op_msg (?P<time_taken>\d+)', line)
                if match:
                    time = logevent.datetime
                    split_range = match.group("range")
                    namespace = match.group("namespace")
                    time_taken = match.group("time_taken")
                    numSplits = 0
                    success = True
                    error = None
                    self._chunk_splits.append((time, split_range, namespace, numSplits, success, time_taken, error))
            elif "Unable to auto-split chunk" in line:
                logevent = LogEvent(line)
                match = re.search("chunk \[(?P<range>.*)\) "
                                  'in namespace (?P<namespace>\S+)'
                                  ' :: caused by :: (?P<error>\S+): ', line)                    
                if match:
                    time = logevent.datetime
                    split_range = match.group("range")
                    namespace = match.group("namespace")
                    numSplits = 0
                    success = False
                    time_taken = 0
                    error = match.group("error")
                    self._chunk_splits.append((time, split_range, namespace, numSplits, success, time_taken, error))
            elif "jumbo" in line:
                logevent = LogEvent(line)
                match = re.search('migration (?P<namespace>\S+): \[(?P<range>.*)\)', prev_line)
                if match:
                    time = logevent.datetime
                    split_range = match.group("range")
                    namespace = match.group("namespace")
                    numSplits = 0
                    success = False
                    time_taken = 0
                    error = "Jumbo"
                    self._chunk_splits.append((time, split_range, namespace, numSplits, success, time_taken, error))
            
            prev_line = line

        # reset logfile
        self.filehandle.seek(0)


    def fast_forward(self, start_dt):
        """
        Fast-forward file to given start_dt datetime obj using binary search.

        Only fast for files. Streams need to be forwarded manually, and it will
        miss the first line that would otherwise match (as it consumes the log
        line).
        """
        if self.from_stdin:
            # skip lines until start_dt is reached
            return

        else:
            # fast bisection path
            max_mark = self.filesize
            step_size = max_mark

            # check if start_dt is already smaller than first datetime
            self.filehandle.seek(0)
            le = self.next()
            if le.datetime and le.datetime >= start_dt:
                self.filehandle.seek(0)
                return

            le = None
            self.filehandle.seek(0)

            # search for lower bound
            while abs(step_size) > 100:
                step_size = ceil(step_size / 2.)

                self.filehandle.seek(step_size, 1)
                le = self._find_curr_line()
                if not le:
                    break

                if le.datetime >= start_dt:
                    step_size = -abs(step_size)
                else:
                    step_size = abs(step_size)

            if not le:
                return

            # now walk backwards until we found a truly smaller line
            while self.filehandle.tell() >= 2 and (le.datetime is None or
                                                   le.datetime >= start_dt):
                self.filehandle.seek(-2, 1)

                le = self._find_curr_line(prev=True)



import re
import json
from collections import defaultdict

from .base_section import BaseSection

try:
    from mtools.util.profile_collection import ProfileCollection
except ImportError:
    ProfileCollection = None


class ClientSection(BaseSection):
    """
    ClientSection class.

    This section goes through the logfile and extracts driver information from
    connected clients.
    """

    name = "clients"

    def __init__(self, mloginfo):
        BaseSection.__init__(self, mloginfo)

        helptext = 'outputs client driver information'
        self.mloginfo.argparser_sectiongroup.add_argument('--clients',
                                                          action='store_true',
                                                          help=helptext)

    @property
    def active(self):
        """Return boolean if this section is active."""
        return(self.mloginfo.args['clients'])

    def run(self):
        """Run this section and print out information."""
        if ProfileCollection and isinstance(self.mloginfo.logfile,
                                            ProfileCollection):
            print("\n    not available for system.profile collections\n")
            return

        """dictionary variable to hold driver name + version as key and each ip
        appended to a list for value against key"""
        driver_info = {}

        for logevent in self.mloginfo.logfile:
            line = logevent.line_str
            pos = line.find('client metadata')
            if pos != -1:

                tokens = line[pos:pos + 100].split(' ')
                ip, _ = tokens[3].split(':')
                ip_formatted = str(ip)

                if ip_formatted != "127.0.0.1":
                    client_metadata = logevent.client_metadata
                    try:
                        driver = client_metadata["driver"]["name"]
                        version = client_metadata["driver"]["version"]
                        dv_formatted = str(driver) + ":" + str(version)
                        if dv_formatted not in driver_info:
                            driver_info[dv_formatted] = [ip_formatted]
                        elif dv_formatted in driver_info:
                            if ip_formatted in driver_info.get(dv_formatted):
                                continue
                            else:
                                driver_info[dv_formatted].append(ip_formatted)
                    except Exception:
                        next

        print('%-15s - Unique connections' % 'Driver:Version ')
        if len(driver_info):
            for key, value in sorted(driver_info.items()):
                print("%-15s : " % (key) + str(value))
        else:
            print("No client metadata found")
        print('')



import re

from .base_filter import BaseFilter

# SERVER-36461 - Filter to log slow transactions; Can be used in conjuncture with other filters


class TransactionFilter(BaseFilter):
    filterArgs = [
        ('--transactions', {
            'action': 'store_true', 'default': False,
            'help': 'only output lines containing logs of transactions'}),
    ]

    def __init__(self, mlogfilter):
        BaseFilter.__init__(self, mlogfilter)

        # set the keyword as transaction if the --transaction flag is set
        if self.mlogfilter.args['transactions']:
            self.words = "transaction"
            self.active = True
        else:
            self.active = False

    def accept(self, logevent):
        """
        Process line.
        Overwrite BaseFilter.accept() and return True if the provided
        logevent contains keyword transaction, or False if not.
        """

        if re.search(self.words, logevent.line_str):
            return True
        else:
            return False



import io
from operator import truediv
import sys

import pytest

from mtools.test import all_tools


@pytest.fixture(scope="function", autouse=True)
def patch_io(monkeypatch, capsys):
    monkeypatch.setattr('sys.stdin', io.StringIO('my input'))
    sys.stdin.name = 'foo'
    sys.stdin.isatty = lambda: True


@all_tools
def check_help(tool_cls):
    """
    Check that all command line tools have a --help option that explains the
    usage.

    As per argparse default, this help text always starts with `usage:`.
    """
    tool = tool_cls()

    try:
        tool.run("--help")

    except SystemExit:
        if not hasattr(sys.stdout, "getvalue"):
            raise Exception('stdout not captured in test.')
        output = sys.stdout.getvalue().strip()
        assert output.startswith('usage:')


@all_tools
def check_version(tool_cls):
    """
    Check that all command line tools have a --version option that returns
    the current version.
    """
    tool = tool_cls()

    try:
        tool.run("--version")

    except SystemExit as ex:
        print(ex)
        if not hasattr(sys.stdout, "getvalue"):
            raise Exception('stdout not captured in test.')

        # argparse's --version outputs to stderr, which can't be captured
        # with pytest. Therefore just checking that the scripts run and not
        # output anything to stdout --- FIXED to stderr https://bugs.python.org/issue18920
        #output = sys.stdout.getvalue().strip()
        #assert len(output) == 0


def test_help():
    check_help()


def test_version():
    check_version()